# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Utilities for deprecating toolkit functionality.
"""

from functools import wraps

from qctrlcommons.exceptions import QctrlException


def deprecate_member(member, deprecated, replacement):
    """
    Replace a member (function or node) in a deprecated namespace
    with a function that raises an exception.
    If a replacement namespace is provided, a suggestion is also given.
    """
    message = f"The {deprecated.value.name} namespace has been deprecated."
    if replacement is not None:
        message += f" Please use the {replacement.value.name} namespace instead."

    @wraps(member)
    def wrapper(*args, **kwargs):  # pylint: disable=unused-argument
        wrapper.__doc__ = message
        raise QctrlException(message)

    return wrapper



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Utility functions for trapped ion systems.
"""
from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass
from typing import (
    Any,
    Optional,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.graph import Graph
from qctrlcommons.node.node_data import Pwc
from qctrlcommons.preconditions import (
    check_argument,
    check_duration,
    check_numeric_numpy_array,
)

from boulderopaltoolkits.namespace import Namespace
from boulderopaltoolkits.toolkit_utils import (
    expose,
    extract_graph_output,
)
from boulderopaltoolkits.utils.nodes import (
    complex_optimizable_pwc_signal,
    real_optimizable_pwc_signal,
)


@expose(Namespace.IONS)
def obtain_ion_chain_properties(
    qctrl: Any,
    atomic_mass: float,
    ion_count: int,
    center_of_mass_frequencies: list,
    wave_numbers: list,
    laser_detuning: Optional[float] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Obtain the Lamb–Dicke parameters and mode frequencies (or relative detunings
    if a laser detuning is provided) for an ion chain.

    This is essentially a wrapper of the Boulder Opal function
    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties`,
    but returns the result as NumPy arrays.

    Parameters
    ----------
    qctrl : qctrl.Qctrl
        Boulder Opal session object.
    atomic_mass : float
        The atomic mass of the ions of the chain in atomic units.
        This function assumes that all the ions in the chain are from the same atomic species.
    ion_count : int
        The number of ions in the chain, :math:`N`.
    center_of_mass_frequencies : list
        A list of three positive numbers representing the center-of-mass trapping frequency
        in the order of radial x-direction, radial y-direction, and axial z-direction,
        which correspond to the unit vectors :math:`(1, 0, 0)`, :math:`(0, 1, 0)`,
        and :math:`(0, 0, 1)` respectively.
    wave_numbers : list
        A list of three elements representing the the laser difference angular wave vector
        (in units of rad/m) in the order of radial x-direction, radial y-direction, and
        the axial z-direction, which correspond to the unit vectors :math:`(1, 0, 0)`,
        :math:`(0, 1, 0)`, and :math:`(0, 0, 1)` respectively.
    laser_detuning : float, optional
        The detuning of the control laser.
        If not provided, the returned relative detunings represent the mode frequencies.

    Returns
    -------
    np.ndarray
        A 3D array of shape ``(3, N, N)`` representing the Lamb–Dicke parameters of the ions.
        Its dimensions indicate, respectively, the direction (radial x-direction,
        radial y-direction, and axial z-direction), the collective mode, and the ion.
    np.ndarray
        A 2D array of shape ``(3, N)`` representing the mode frequencies
        (or relative detunings if a laser detuning is provided).
        Its dimensions indicate, respectively, the direction (radial x-direction,
        radial y-direction, and axial z-direction) and the collective mode.

    See Also
    --------
    :func:`~qctrl.dynamic.namespaces.FunctionNamespace.calculate_ion_chain_properties` :
        Function to calculate the properties of an ion chain.
    :func:`.ions.ms_optimize` :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    :func:`.ions.ms_simulate` :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.

    Examples
    --------
    Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust
    -molmer-sorensen-gates-for-trapped-ions>`_ user guide to find how to use this function.
    """

    check_argument(
        isinstance(ion_count, (int, np.integer)) and ion_count >= 1,
        "The ion count must be an integer and greater than 0.",
        {"ion_count": ion_count},
    )
    check_argument(
        _is_positive(atomic_mass),
        "The atomic mass must be positive.",
        {"atomic_mass": atomic_mass},
    )
    check_argument(
        len(center_of_mass_frequencies) == 3,
        "The center_of_mass_frequencies list must have three elements, representing the frequency "
        "in the order of radial x-direction, radial y-direction, and the axial z-direction.",
        {"center_of_mass_frequencies": center_of_mass_frequencies},
    )
    check_argument(
        _is_positive(np.asarray(center_of_mass_frequencies)),
        "All center of mass frequencies must be positive.",
        {"center_of_mass_frequencies": center_of_mass_frequencies},
    )
    check_argument(
        len(wave_numbers) == 3,
        "The wave_numbers list must have three elements, representing the wave vector "
        "in the order of radial x-direction, radial y-direction, and the axial z-direction.",
        {"wave_numbers": wave_numbers},
    )
    check_argument(
        not np.allclose(wave_numbers, 0),
        "At least one of the wave numbers must be non-zero.",
        {"wave_numbers": wave_numbers},
    )

    ion_chain_properties = qctrl.functions.calculate_ion_chain_properties(
        atomic_mass=float(atomic_mass),
        ion_count=int(ion_count),
        radial_x_center_of_mass_frequency=float(center_of_mass_frequencies[0]),
        radial_y_center_of_mass_frequency=float(center_of_mass_frequencies[1]),
        axial_center_of_mass_frequency=float(center_of_mass_frequencies[2]),
        radial_x_wave_number=float(wave_numbers[0]),
        radial_y_wave_number=float(wave_numbers[1]),
        axial_wave_number=float(wave_numbers[2]),
    )

    # The first axis of lamb_dicke_parameters and detuning are directions (x, y, z).
    lamb_dicke_parameters = []
    frequencies = []

    for properties_per_direction in [
        ion_chain_properties.radial_x_mode_properties,
        ion_chain_properties.radial_y_mode_properties,
        ion_chain_properties.axial_mode_properties,
    ]:
        lamb_dicke_parameters.append(
            [p.lamb_dicke_parameters for p in properties_per_direction]
        )
        frequencies.append([p.frequency for p in properties_per_direction])

    if laser_detuning is None:
        return np.asarray(lamb_dicke_parameters), np.asarray(frequencies)

    return np.asarray(lamb_dicke_parameters), np.asarray(frequencies) - laser_detuning


def _is_positive(x: float | np.ndarray | int):
    """
    Check the value must be greater than zero, taking into account rounding errors.
    """
    return not np.any(np.isclose(x, 0)) and np.all(x > 0)


def _complex_symmetrized_drive(graph, segment_count, duration, maximum, name):
    """
    Create a symmetrized drive signal (Milne et al., Phys. Rev. Applied, 2020).
    """

    free_segment_count = (segment_count + 1) // 2

    moduli = graph.optimization_variable(
        count=free_segment_count, lower_bound=0, upper_bound=maximum
    )
    phases = graph.optimization_variable(
        count=free_segment_count,
        lower_bound=0,
        upper_bound=2 * np.pi,
        is_lower_unbounded=True,
        is_upper_unbounded=True,
    )

    if segment_count % 2 == 0:
        moduli_reversed = graph.reverse(moduli, [0])
        phases_reversed = graph.reverse(phases, [0])

    else:
        moduli_reversed = graph.reverse(moduli[:-1], [0])
        phases_reversed = graph.reverse(phases[:-1], [0])

    moduli_comb = graph.concatenate([moduli, moduli_reversed], 0)
    phases_comb = graph.concatenate([phases, 2 * phases[-1] - phases_reversed], 0)

    return graph.complex_pwc_signal(
        moduli=moduli_comb, phases=phases_comb, duration=duration, name=name
    )


def _real_symmetrized_drive(graph, segment_count, duration, minimum, maximum, name):
    """
    Create a symmetrized drive signal (Milne et al., Phys. Rev. Applied, 2020).
    """

    free_segment_count = (segment_count + 1) // 2

    values = graph.optimization_variable(
        count=free_segment_count, lower_bound=minimum, upper_bound=maximum
    )
    if segment_count % 2 == 0:
        values_reversed = graph.reverse(values, [0])

    else:
        values_reversed = graph.reverse(values[:-1], [0])

    return graph.pwc_signal(
        values=graph.concatenate([values, values_reversed], 0),
        duration=duration,
        name=name,
    )


def _validate_addressing(addressing):
    """
    Validate an addressing input and return it as a tuple.
    """
    message = "The ions addressed must be an integer or a tuple of integers."
    try:
        check_argument(
            all(isinstance(ion, (int, np.integer)) for ion in addressing),
            message,
            {"addressing": addressing},
        )

    except TypeError as error:
        if isinstance(addressing, (int, np.integer)):
            return (addressing,)

        raise QctrlArgumentsValueError(message, {"addressing": addressing}) from error

    addressing = tuple(int(k) for k in addressing)
    check_argument(
        len(addressing) == len(set(addressing)),
        "The ions addressed must be unique.",
        {"addressing": addressing},
    )
    return addressing


@expose(Namespace.IONS)
@dataclass
class Drive:
    """
    A piecewise-constant complex-valued drive.

    Parameters
    ----------
    values : np.ndarray
        The values of the drive at each segment, in units of rad/s.
    addressing : int or tuple[int, ...]
        The indices of the ions addressed by the drive.

    See Also
    --------
    :class:`.ions.OptimizableDrive` :
        Abstract class describing a piecewise-constant optimizable drive.
    :func:`.ions.ms_simulate` :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.
    """

    values: np.ndarray
    addressing: int | tuple[int, ...]

    def __post_init__(self):
        check_numeric_numpy_array(self.values, "values")
        self.addressing = _validate_addressing(self.addressing)

    def get_pwc(self, graph: Graph, duration: float) -> Pwc:
        """
        Return a Pwc representation of the drive.
        """
        return graph.pwc_signal(values=self.values, duration=duration)


@expose(Namespace.IONS)
class OptimizableDrive(ABC):
    """
    Abstract class for optimizable drives. You need to call the concrete classes below
    to create optimizable drives.

    See Also
    --------
    :class:`.ions.ComplexOptimizableDrive` :
        Class describing a piecewise-constant complex-valued optimizable drive.
    :class:`.ions.RealOptimizableDrive` :
        Class describing a piecewise-constant real-valued optimizable drive.
    :func:`.ions.ms_optimize` :
       Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    """

    name: str
    addressing: int | tuple[int, ...]

    @abstractmethod
    def get_pwc(self, graph: Graph, duration: float, robust: bool) -> Pwc:
        """
        Return a Pwc representation of the optimizable drive.
        """
        raise NotImplementedError


@expose(Namespace.IONS)
@dataclass
class ComplexOptimizableDrive(OptimizableDrive):
    """
    A piecewise-constant complex-valued optimizable drive.
    The main function will try to find the optimal values for it.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant drive.
    maximum_rabi_rate : float
        The maximum value that the modulus of the drive can take at each segment,
        in units of rad/s.
    addressing : int or tuple[int, ...]
        The indices of the ions addressed by the drive.
    name : str, optional
        The identifier of the drive.
        Defaults to "drive".

    See Also
    --------
    :class:`.ions.Drive` :
        Class describing non-optimizable drives.
    :class:`.ions.RealOptimizableDrive` :
        Class describing optimizable real-valued drives.
    :func:`.ions.ms_optimize` :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    """

    count: int
    maximum_rabi_rate: float
    addressing: int | tuple[int, ...]
    name: str = "drive"

    def __post_init__(self):
        check_argument(
            self.count > 0, "There must be at least one segment.", {"count": self.count}
        )
        self.addressing = _validate_addressing(self.addressing)

    def get_pwc(self, graph: Graph, duration: float, robust: bool) -> Pwc:
        if robust:
            return _complex_symmetrized_drive(
                graph=graph,
                segment_count=self.count,
                duration=duration,
                maximum=self.maximum_rabi_rate,
                name=self.name,
            )

        return complex_optimizable_pwc_signal(
            graph=graph,
            segment_count=self.count,
            duration=duration,
            maximum=self.maximum_rabi_rate,
            name=self.name,
        )


@expose(Namespace.IONS)
@dataclass
class RealOptimizableDrive(OptimizableDrive):
    """
    A piecewise-constant real-valued optimizable drive.
    The main function will try to find the optimal values for it.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant drive.
    minimum_rabi_rate : float
        The minimum value that the drive can take at each segment, in units of rad/s.
    maximum_rabi_rate : float
        The maximum value that the drive can take at each segment, in units of rad/s.
    addressing : int or tuple[int, ...]
        The indices of the ions addressed by the drive.
    name : str, optional
        The identifier of the drive.
        Defaults to "drive".

    See Also
    --------
    :class:`.ions.ComplexOptimizableDrive` :
        Class describing optimizable complex-valued drives.
    :class:`.ions.Drive` :
        Class describing non-optimizable drives.
    :func:`.ions.ms_optimize` :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    """

    count: int
    minimum_rabi_rate: float
    maximum_rabi_rate: float
    addressing: int | tuple[int, ...]
    name: str = "drive"

    def __post_init__(self):
        check_argument(
            self.count > 0, "There must be at least one segment.", {"count": self.count}
        )
        check_argument(
            self.minimum_rabi_rate < self.maximum_rabi_rate,
            "The maximum Rabi rate must be larger than the minimum.",
            {
                "minimum_rabi_rate": self.minimum_rabi_rate,
                "maximum_rabi_rate": self.maximum_rabi_rate,
            },
        )
        self.addressing = _validate_addressing(self.addressing)

    def get_pwc(self, graph: Graph, duration: float, robust: bool) -> Pwc:
        if robust:
            return _real_symmetrized_drive(
                graph=graph,
                segment_count=self.count,
                duration=duration,
                minimum=self.minimum_rabi_rate,
                maximum=self.maximum_rabi_rate,
                name=self.name,
            )

        return real_optimizable_pwc_signal(
            graph=graph,
            segment_count=self.count,
            duration=duration,
            maximum=self.maximum_rabi_rate,
            minimum=self.minimum_rabi_rate,
            name=self.name,
        )


def _validate_system_parameters(
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    target_phases: Optional[np.ndarray],
) -> int:
    """
    Validate the arrays describing an ion system
    and return the number of ions.
    """

    ion_count = lamb_dicke_parameters.shape[-1]

    check_argument(
        lamb_dicke_parameters.shape == (3, ion_count, ion_count)
        and relative_detunings.shape == (3, ion_count),
        "The shape of the Lamb–Dicke parameters array must be (3, N, N), "
        "and the shape of the relative detunings array must be (3, N), "
        "where N is the number of ions.",
        {
            "lamb_dicke_parameters": lamb_dicke_parameters,
            "relative_detunings": relative_detunings,
        },
        extras={
            "lamb_dicke_parameters.shape": lamb_dicke_parameters.shape,
            "relative_detunings.shape": relative_detunings.shape,
        },
    )

    if target_phases is not None:
        check_argument(
            target_phases.shape == (ion_count, ion_count),
            "The shape of the target phases array must be (N, N), "
            "where N is the number of ions.",
            {"target_phases": target_phases},
            extras={"ion count": ion_count, "target_phases.shape": target_phases.shape},
        )

    return ion_count


def _check_drives_addressing(drives, ion_count):
    """
    Check the input drives are a list and that the ions they address are valid.
    """

    check_argument(
        isinstance(drives, list),
        "You must provide a list of drives.",
        {"drives": drives},
    )

    all_addressing = []
    for idx, drive in enumerate(drives):
        check_argument(
            all(0 <= ion < ion_count for ion in drive.addressing),
            "The addressed ions must be between 0 (inclusive) "
            "and the number of ions (exclusive).",
            {"drives": drives},
            extras={
                f"drives[{idx}].addressing": drive.addressing,
                "ion count": ion_count,
            },
        )
        all_addressing += drive.addressing

    check_argument(
        len(all_addressing) == len(set(all_addressing)),
        "Each ion can only be addressed by a single drive.",
        {"drives": drives},
    )


def _get_ion_drives(pwc_addressings, ion_count, graph, duration):
    """
    From a list of (Pwc, list(int)) tuples (drives and ions addressed by them),
    return a list of length ion_count the drive addressing each ion
    or a PWC with value 0 if the ion is not addressed by any drive.
    """
    ion_drives = []
    for idx in range(ion_count):
        for pwc, addressing in pwc_addressings:
            # Add the first drive that addresses the ion as we assume
            # each ion can only be addressed by a single drive.
            if idx in addressing:
                ion_drives.append(pwc)
                break
        else:
            ion_drives.append(graph.constant_pwc(constant=0.0, duration=duration))

    return ion_drives


_MS_NODE_NAMES = ["phases", "displacements", "infidelities"]


@expose(Namespace.IONS)
def ms_simulate(
    qctrl: Any,
    drives: list[Drive],
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    duration: float,
    target_phases: Optional[np.ndarray] = None,
    sample_count: int = 128,
):
    r"""
    Simulate a Mølmer–Sørensen-type operation on a system composed of ions.

    Parameters
    ----------
    qctrl : qctrl.Qctrl
        Boulder Opal session object.
    drives : list[~ions.Drive]
        A list of drives addressing the ions.
        Each ion can only be addressed by a single drive,
        but there may be ions not addressed by any drive.
    lamb_dicke_parameters : np.ndarray
        A 3D array of shape ``(3, N, N)``, where :math:`N` is the number of
        ions in the system, specifying the laser-ion coupling strength,
        :math:`\{\eta_{jkl}\}`. The three dimensions indicate, respectively,
        the axis, the collective mode number, and the ion.
    relative_detunings : np.ndarray
        A 2D array of shape ``(3, N)`` specifying the difference, in Hz, between
        each motional mode frequency and the laser detuning
        (with respect to the qubit transition frequency :math:`\omega_0`),
        :math:`\{\delta_{jk} = \nu_{jk} - \delta\}`. The two dimensions indicate,
        respectively, the axis and the collective mode number.
    duration : float
        The duration, in seconds, of the dynamics to be simulated, :math:`T`.
        It must be greater than zero.
    target_phases : np.ndarray, optional
        A 2D array of shape ``(N, N)`` with the target relative phases between
        ion pairs, :math:`\{\Psi_{kl}\}`, as a strictly lower triangular matrix.
        Its :math:`(k, l)`-th element indicates the total relative phase target
        for ions :math:`k` and :math:`l`, with :math:`k > l`.
        If not provided, the function does not return the operational infidelities.
    sample_count : int, optional
        The number of times :math:`T` between 0 and `duration` (included)
        at which the evolution is sampled.
        Defaults to 128.

    Returns
    -------
    dict
        A dictionary containing information about the evolution of the system.
        The dictionary keys are:

            ``sample_times``
                The times at which the evolution is sampled, as an array of shape ``(T,)``.
            ``phases``
                Acquired phases :math:`\{\Phi_{jk}(t_i) = \phi_{jk}(t_i) + \phi_{kj}(t_i)\}`
                for each sample time and for all ion pairs, as a strictly lower triangular
                matrix of shape ``(T, N, N)``.
                :math:`\Phi_{jk}` records the relative phase between ions :math:`j`
                and :math:`k`; matrix elements where :math:`j \leq k` are zero.
            ``displacements``
                Displacements :math:`\{\eta_{pj}\alpha_{pj}(t_i)\}` for all mode-ion
                combinations, as an array of shape ``(T, 3, N, N)``.
                The four dimensions indicate the sample time, the axis,
                the collective mode number, and the ion.
            ``infidelities``
                A 1D array of length ``T`` representing the operational infidelities of
                the Mølmer–Sørensen gate at each sample time, :math:`\mathcal{I}(t_i)`.
                Only returned if target relative phases are provided.

    See Also
    --------
    :class:`.ions.Drive` :
        Class describing non-optimizable drives.
    :func:`.ions.ms_optimize` :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    :func:`.ions.obtain_ion_chain_properties` :
        Calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{p = 1}^{3N} \hbar\nu_p \left(a_p^\dagger a_p + \frac{1}{2}\right)
            + \sum_{j = 1}^N \frac{\hbar \omega_0}{2} \sigma_{z,j} ,

    where the axis dimension and collective mode dimension are combined into a single index
    :math:`p` for simplicity, :math:`a_p` is the annihilation operator for the mode :math:`p`,
    and :math:`\sigma_{z,j}` is the Pauli :math:`Z` operator for the ion :math:`j`.
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} (-\beta_{pj}^*(t)a_p +
                \beta_{pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{pj}(t) = \eta_{pj} \frac{\gamma_j(t)}{2} e^{i\delta_p t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`,
    where :math:`\{\gamma_j\}` is the total drive acting on ion :math:`j`.

    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N\sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N}  \left(\eta_{pj}\alpha_{pj}(t)a_p^\dagger
            - \eta_{pj}^{\ast}\alpha_{pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \int_{0}^{t} d \tau_1
            \int_{0}^{\tau_1} d \tau_2 \beta_{pj}(\tau_1)\beta_{pk}^{\ast}(\tau_2) \right] ,

    and

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2} e^{i \delta_p \tau} .

    The operational infidelity of the Mølmer–Sørensen gate is defined as [1]_:

    .. math::
        \mathcal{I} = 1 - \left| \left( \prod_{\substack{k=1 \\ l<k}}^N \cos (
            \phi_{kl} - \psi_{kl}) \right)
            \left( 1 - \sum_{j=1}^3 \sum_{k,l=1}^N \left[ |\eta_{jkl}|^2
            |\alpha_{jkl}|^2 \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 .

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_
    """

    ion_count = _validate_system_parameters(
        lamb_dicke_parameters, relative_detunings, target_phases
    )

    check_duration(duration, "duration")

    check_argument(
        all(isinstance(drive, Drive) for drive in drives),
        "All drives must be non-optimizable.",
        {"drives": drives},
    )

    sample_times = np.linspace(0.0, duration, sample_count)

    graph = qctrl.create_graph()

    _check_drives_addressing(drives, ion_count)

    drive_pwcs = [
        (drive.get_pwc(graph, duration), drive.addressing) for drive in drives
    ]
    ion_drives = _get_ion_drives(drive_pwcs, ion_count, graph, duration)

    phases = graph.ms_phases(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[0],
    )

    displacements = graph.ms_displacements(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[1],
    )

    if target_phases is not None:
        graph.ms_infidelity(
            phases=phases,
            displacements=displacements,
            target_phases=target_phases,
            name=_MS_NODE_NAMES[2],
        )
        output_node_names = _MS_NODE_NAMES
    else:
        output_node_names = _MS_NODE_NAMES[:2]

    simulation_result = qctrl.functions.calculate_graph(
        graph=graph, output_node_names=output_node_names
    )

    # Retrieve results and build output dictionary.
    result_dict: dict[str, Any] = {"sample_times": sample_times}

    for key, value in simulation_result.output.items():
        result_dict[key] = extract_graph_output(value)

    return result_dict


@expose(Namespace.IONS)
def ms_optimize(
    qctrl: Any,
    drives: list[OptimizableDrive],
    lamb_dicke_parameters: np.ndarray,
    relative_detunings: np.ndarray,
    duration: float,
    target_phases: np.ndarray,
    sample_count: int = 128,
    robust: bool = False,
    optimization_count: int = 4,
    max_iteration_count: Optional[int] = None,
    cost_history: Optional[str] = None,
):
    r"""
    Find optimal pulses to perform a target Mølmer–Sørensen-type operation
    on a system composed of ions.

    Parameters
    ----------
    qctrl : qctrl.Qctrl
        Boulder Opal session object.
    drives : list[OptimizableDrive]
        A list of optimizable drives addressing the ions.
        Each ion can only be addressed by a single drive,
        but there may be ions not addressed by any drive.
    lamb_dicke_parameters : np.ndarray
        A 3D array of shape ``(3, N, N)``, where :math:`N` is the number of
        ions in the system, specifying the laser-ion coupling strength,
        :math:`\{\eta_{jkl}\}`. The three dimensions indicate, respectively,
        the axis, the collective mode number, and the ion.
    relative_detunings : np.ndarray
        A 2D array of shape ``(3, N)`` specifying the difference, in Hz, between
        each motional mode frequency and the laser detuning
        (with respect to the qubit transition frequency :math:`\omega_0`),
        :math:`\{\delta_{jk} = \nu_{jk} - \delta\}`. The two dimensions indicate,
        respectively, the axis and the collective mode number.
    duration : float
        The duration, in seconds, of the dynamics to be optimized, :math:`T`.
        It must be greater than zero.
    target_phases : np.ndarray
        A 2D array of shape ``(N, N)`` with the target relative phases between
        ion pairs, :math:`\{\Psi_{kl}\}`, as a strictly lower triangular matrix.
        Its :math:`(k, l)`-th element indicates the total relative phase target
        for ions :math:`k` and :math:`l`, with :math:`k > l`.
    sample_count : int, optional
        The number of times :math:`T` between 0 and `duration` (included)
        at which the evolution is sampled.
        Defaults to 128.
    robust : bool, optional
        Whether to obtain optimized drives robust against dephasing noise.
    optimization_count : int, optional
        The number of independent randomly seeded optimizations to perform. The result
        from the best optimization (the one with the lowest cost) is returned.
        Defaults to 4.
    max_iteration_count : int, optional
        The maximum number of optimizer iterations to perform.
        You can set this as an early stop condition for the optimizer.
        If provided, it must be greater than zero.
        Defaults to None, which means that the optimizer runs until it converges.
    cost_history : str, optional
        String determining the output of the optimization cost history.
        If "BEST", a single cost history for the best performing optimization is returned.
        If "ALL", a list with the cost histories for all optimizations is returned.
        Defaults to None, in which case no cost history is returned.

    Returns
    -------
    dict
        A dictionary containing information about the evolution of the system.
        The dictionary keys are:

            ``optimized_values``
                Dictionary containing the optimized drives values, whose keys
                are the names of the drives provided to the function, and whose
                values are an np.ndarray with the values of the piecewise-constant signal.
            ``cost``
                The final optimized cost. If `robust` is set to False, this corresponds to
                the infidelity at the end of the gate. If `robust` is set to True, it is
                the final infidelity plus the dephasing-robust cost term.
            ``sample_times``
                The times at which the evolution is sampled, as an array of shape ``(T,)``.
            ``phases``
                Acquired phases :math:`\{\Phi_{jk}(t_i) = \phi_{jk}(t_i) + \phi_{kj}(t_i)\}`
                for each sample time and for all ion pairs, as a strictly lower triangular
                matrix of shape ``(T, N, N)``.
                :math:`\Phi_{jk}` records the relative phase between ions :math:`j`
                and :math:`k`; matrix elements where :math:`j \leq k` are zero.
            ``displacements``
                Displacements :math:`\{\eta_{pj}\alpha_{pj}(t_i)\}` for all mode-ion
                combinations, as an array of shape ``(T, 3, N, N)``.
                The four dimensions indicate the sample time, the axis,
                the collective mode number, and the ion.
            ``infidelities``
                A 1D array of length ``T`` representing the operational infidelities of
                the Mølmer–Sørensen gate at each sample time, :math:`\mathcal{I}(t_i)`.
            ``cost_history``
                If `cost_history` is set to "BEST", the cost history for the best performing
                optimization. If `cost_history` is set to "ALL", a list of all the cost histories
                for each optimization.

    See Also
    --------
    :class:`.ions.ComplexOptimizableDrive` :
        Class describing a piecewise-constant complex-valued optimizable drive.
    :class:`.ions.RealOptimizableDrive` :
        Class describing a piecewise-constant real-valued optimizable drive.
    :func:`.ions.ms_simulate` :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.
    :func:`.ions.obtain_ion_chain_properties` :
        Calculate the properties of an ion chain.

    Notes
    -----
    The internal and motional Hamiltonian of :math:`N` ions is

    .. math::
        H_0 = \sum_{p = 1}^{3N} \hbar\nu_p \left(a_p^\dagger a_p + \frac{1}{2}\right)
            + \sum_{j = 1}^N \frac{\hbar \omega_0}{2} \sigma_{z,j} ,

    where the axis dimension and collective mode dimension are combined into a single index
    :math:`p` for simplicity, :math:`a_p` is the annihilation operator for the mode :math:`p`,
    and :math:`\sigma_{z,j}` is the Pauli :math:`Z` operator for the ion :math:`j`.
    The interaction Hamiltonian for Mølmer–Sørensen-type
    operations in the rotating frame with respect to :math:`H_0` is:

    .. math::
        H_I(t) = i\hbar\sum_{j = 1}^N \sigma_{x, j} \sum_{p = 1}^{3N} (-\beta_{pj}^*(t)a_p +
                \beta_{pj}(t) a_p^\dagger) ,

    where :math:`\sigma_{x, j}` is the Pauli :math:`X` operator for the ion :math:`j` and
    :math:`\beta_{pj}(t) = \eta_{pj} \frac{\gamma_j(t)}{2} e^{i\delta_p t}`,
    indicating the coupling of the ion :math:`j` to the motional mode :math:`p`,
    where :math:`\{\gamma_j\}` is the total drive acting on ion :math:`j`.

    The corresponding unitary operation is given by [1]_

    .. math::
        U(t) = \exp\left[ \sum_{j=1}^N \sigma_{x, j} B_j(t)
                + i\sum_{j=1}^N\sum_{k=1}^{j - 1} (\phi_{jk}(t) + \phi_{kj}(t))
                \sigma_{x, j} \sigma_{x, k} \right] ,

    where

    .. math::
        B_j(t) &\equiv \sum_{p = 1}^{3N}  \left(\eta_{pj}\alpha_{pj}(t)a_p^\dagger
            - \eta_{pj}^{\ast}\alpha_{pj}^\ast(t)a_p \right) ,

        \phi_{jk}(t) &\equiv \mathrm{Im} \left[ \sum_{p=1}^{3N} \int_{0}^{t} d \tau_1
            \int_{0}^{\tau_1} d \tau_2 \beta_{pj}(\tau_1)\beta_{pk}^{\ast}(\tau_2) \right] ,

    and

    .. math::
        \alpha_{pj}(t) = \int_0^t d\tau \frac{\gamma_j(\tau)}{2} e^{i \delta_p \tau} .

    The operational infidelity of the Mølmer–Sørensen gate is defined as [1]_:

    .. math::
        \mathcal{I} = 1 - \left| \left( \prod_{\substack{k=1 \\ l<k}}^N \cos (
            \phi_{kl} - \psi_{kl}) \right)
            \left( 1 - \sum_{j=1}^3 \sum_{k,l=1}^N \left[ |\eta_{jkl}|^2
            |\alpha_{jkl}|^2 \left(\bar{n}_{jk}+\frac{1}{2} \right) \right] \right) \right|^2 .

    You can use the `robust` flag to construct a Mølmer–Sørensen gate that is
    robust against dephasing noise. This imposes a symmetry [1]_ in the optimizable
    ion drives and aims to minimize the time-averaged positions of the phase-space
    trajectories,

    .. math::
        \langle \alpha_{pj} \rangle
            = \frac{1}{t_\text{gate}} \int_0^{t_\text{gate}}
                \alpha_{pj}(t) \mathrm{d} t ,

    where the axis dimension and the collective mode dimension are combined
    into a single index :math:`p` for simplicity.

    This is achieved by adding an additional term to the cost function,
    consisting of the sum of the square moduli of the time-averaged positions
    multiplied by the corresponding Lamb–Dicke parameters. That is to say,

    .. math::
        C_\text{robust} =
            \mathcal{I} + \sum_{p,j}
                \left| \eta_{pj} \langle \alpha_{pj} \rangle \right|^2 .

    References
    ----------
    .. [1] `C. D. B. Bentley, H. Ball, M. J. Biercuk, A. R. R. Carvalho,
            M. R. Hush, and H. J. Slatyer, Advanced Quantum Technologies 3, 2000044 (2020).
            <https://doi.org/10.1002/qute.202000044>`_
    """

    if max_iteration_count is not None:
        check_argument(
            max_iteration_count > 0,
            "The maximum number of iterations must be positive.",
            {"max_iteration_count": max_iteration_count},
        )

    ion_count = _validate_system_parameters(
        lamb_dicke_parameters, relative_detunings, target_phases
    )

    check_duration(duration, "duration")

    _check_drives_addressing(drives, ion_count)

    check_argument(
        all(isinstance(drive, OptimizableDrive) for drive in drives),
        "All drives must be optimizable.",
        {"drives": drives},
    )

    sample_times = np.linspace(0.0, duration, sample_count)

    graph = qctrl.create_graph()

    drive_names = [drive.name for drive in drives]

    check_argument(
        len(drive_names) == len(set(drive_names)),
        "The drive names must be unique.",
        {"drives": drives},
        extras={"[drive.name for drive in drives]": drive_names},
    )

    if cost_history:
        check_argument(
            cost_history in ["ALL", "BEST"],
            "If passed, the cost history must be 'ALL' or 'BEST'.",
            {"cost_history": cost_history},
        )
        cost_history_scope = "ITERATION_VALUES"
    else:
        cost_history_scope = None

    drive_pwcs = [
        (drive.get_pwc(graph, duration, robust), drive.addressing) for drive in drives
    ]
    ion_drives = _get_ion_drives(drive_pwcs, ion_count, graph, duration)

    phases = graph.ms_phases(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[0],
    )

    displacements = graph.ms_displacements(
        drives=ion_drives,
        lamb_dicke_parameters=lamb_dicke_parameters,
        relative_detunings=relative_detunings,
        sample_times=sample_times,
        name=_MS_NODE_NAMES[1],
    )

    infidelities = graph.ms_infidelity(
        phases=phases,
        displacements=displacements,
        target_phases=target_phases,
        name=_MS_NODE_NAMES[2],
    )

    cost = infidelities[-1]
    if robust:
        cost += graph.ms_dephasing_robust_cost(
            drives=ion_drives,
            lamb_dicke_parameters=lamb_dicke_parameters,
            relative_detunings=relative_detunings,
        )

    optimization_result = qctrl.functions.calculate_optimization(
        graph=graph,
        optimization_count=optimization_count,
        cost_node_name=cost.name,
        output_node_names=drive_names + _MS_NODE_NAMES,
        max_iteration_count=max_iteration_count,
        cost_history_scope=cost_history_scope,
    )

    # Retrieve results and build output dictionary.
    result_dict: dict[str, Any] = {
        "optimized_values": {},
        "cost": optimization_result.cost,
        "sample_times": sample_times,
    }

    for key in drive_names:
        result_dict["optimized_values"][key] = extract_graph_output(
            optimization_result.output[key]
        )

    for key in _MS_NODE_NAMES:
        result_dict[key] = extract_graph_output(optimization_result.output[key])

    if cost_history == "ALL":
        result_dict["cost_history"] = optimization_result.cost_history.iteration_values
    elif cost_history == "BEST":
        result_dict["cost_history"] = min(
            optimization_result.cost_history.iteration_values, key=lambda hist: hist[-1]
        )

    return result_dict



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
System-agnostic convenient nodes.
"""

from __future__ import annotations

from typing import Optional

import numpy as np
from qctrlcommons.graph import Graph
from qctrlcommons.node.node_data import Pwc
from qctrlcommons.preconditions import check_argument

from boulderopaltoolkits.namespace import Namespace
from boulderopaltoolkits.toolkit_utils import expose


@expose(Namespace.UTILS)
def real_optimizable_pwc_signal(
    graph: Graph,
    segment_count: int,
    duration: float,
    maximum: float,
    minimum: float = 0.0,
    initial_values: Optional[np.ndarray | list[np.ndarray]] = None,
    name: Optional[str] = None,
) -> Pwc:
    """
    Create a real optimizable piecewise-constant signal.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    segment_count : int
        The number of piecewise-constant segments in the signal.
    duration : float
        The duration of the signal.
    maximum : float
        The upper bound for the signal values.
    minimum : float, optional
        The lower bound for the signal values. Defaults to 0.
    initial_values : np.ndarray or list[np.ndarray], optional
        Initial values for the signal. Defaults to None, meaning the optimizer
        initializes these variables with random values. You can either provide a single array,
        or a list of them. If a list of arrays are used, they must have the same length.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The optimizable piecewise-constant signal.

    See Also
    --------
    :func:`.utils.complex_optimizable_pwc_signal` :
        Create a complex optimizable `Pwc` signal.
    :func:`~qctrl.graphs.Graph.optimization_variable` :
        Create optimization variables, which can be bounded, semi-bounded, or unbounded.
    :func:`~qctrl.graphs.Graph.pwc_signal` : Create a piecewise-constant signal.
    """

    if initial_values is not None:
        check_argument(
            np.all(np.isreal(initial_values)),
            "Initial signal values must be real.",
            {"initial_values": initial_values},
        )

    values = graph.optimization_variable(
        count=segment_count,
        lower_bound=minimum,
        upper_bound=maximum,
        initial_values=initial_values,
    )
    return graph.pwc_signal(values=values, duration=duration, name=name)


@expose(Namespace.UTILS)
def complex_optimizable_pwc_signal(
    graph: Graph,
    segment_count: int,
    duration: float,
    maximum: float,
    initial_values: Optional[np.ndarray | list[np.ndarray]] = None,
    name: Optional[str] = None,
) -> Pwc:
    """
    Create a complex optimizable piecewise-constant signal.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    segment_count : int
        The number of segments of the signal.
    duration : float
        The duration of the signal.
    maximum : float
        The upper bound for the modulus of the signal values.
    initial_values : np.ndarray or list[np.ndarray], optional
        Initial values for the signal. Defaults to None, meaning the optimizer
        initializes these variables with random values. You can either provide a single array,
        or a list of them. If a list of arrays are used, they must have the same length.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The optimizable piecewise-constant signal.

    See Also
    --------
    :func:`.utils.real_optimizable_pwc_signal` :
        Create a real optimizable `Pwc` signal.
    :func:`~qctrl.graphs.Graph.complex_pwc_signal` :
        Create a complex piecewise-constant signal from moduli and phases.
    :func:`~qctrl.graphs.Graph.optimization_variable` :
        Create optimization variables, which can be bounded, semi-bounded, or unbounded.
    :func:`~qctrl.graphs.Graph.pwc_signal` : Create a piecewise-constant signal.

    Notes
    -----
    Note that this function sets limits to the modulus of the signal.

    If you want to set (different) limits to the real and imaginary parts instead,
    consider using `graph.utils.real_optimizable_signal` to create signals for the
    real and imaginary parts, which you can pass to `graph.complex_value`.
    """

    initial_moduli: Optional[list[np.ndarray] | np.ndarray] = None
    initial_phases: Optional[list[np.ndarray] | np.ndarray] = None

    if initial_values is not None:
        if isinstance(initial_values, list):
            initial_moduli = list(np.absolute(initial_values))
            initial_phases = list(np.angle(initial_values))
        else:
            initial_moduli = np.absolute(initial_values)
            initial_phases = np.angle(initial_values)

    moduli = graph.optimization_variable(
        count=segment_count,
        lower_bound=0.0,
        upper_bound=maximum,
        initial_values=initial_moduli,
    )
    phases = graph.optimization_variable(
        count=segment_count,
        lower_bound=-np.pi,
        upper_bound=np.pi,
        is_lower_unbounded=True,
        is_upper_unbounded=True,
        initial_values=initial_phases,
    )
    return graph.complex_pwc_signal(
        moduli=moduli, phases=phases, duration=duration, name=name
    )


@expose(Namespace.UTILS)
def filter_and_resample_pwc(
    graph: Graph,
    pwc: Pwc,
    cutoff_frequency: float,
    segment_count: int,
    duration: Optional[float] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Filter a piecewise-constant function with a sinc filter and resample it again.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    pwc : Pwc
        The piecewise-constant function :math:`\alpha(t)` to be filtered.
    cutoff_frequency : float
        Upper limit :math:`\omega_c` of the range of angular frequencies that you want to
        preserve in your function.
    segment_count : int
        The number of segments of the resampled filtered function.
    duration : float, optional
        Force the resulting piecewise-constant function to have a certain duration.
        This option is mainly to avoid floating point errors when the total duration is
        too small. If not provided, it is set to the sum of segment durations of `pwc`.
        Defaults to None.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The filtered and resampled piecewise-constant function.

    See Also
    --------
    :func:`~qctrl.graphs.Graph.convolve_pwc` :
        Create the convolution of a piecewise-constant function with a kernel.
    :func:`~qctrl.graphs.Graph.discretize_stf` :
        Create a piecewise-constant function by discretizing a sampleable function.
    :func:`~qctrl.graphs.Graph.sinc_convolution_kernel` :
        Create a convolution kernel representing the sinc function.

    Notes
    -----
    The resulting filtered function is

    .. math::
        \int_{-\infty}^\infty \alpha(\tau)
        \frac{\sin[\omega_c (t-\tau)]}{\pi (t-\tau)} \mathrm{d}\tau
        = \frac{1}{2\pi} \int_{-\omega_c}^{\omega_c}
        e^{i \omega t} \hat\alpha(\omega) \mathrm{d}\omega

    where

    .. math::
        \hat\alpha(\omega)
        =\int_{-\infty}^\infty e^{-i \omega \tau}\alpha(\tau) \mathrm{d}\tau

    is the Fourier transform of :math:`\alpha(t)`. Hence the filter eliminates components of
    the signal that have angular frequencies greater than :math:`\omega_c`.
    """
    if duration is not None:
        check_argument(
            duration > 0,
            "The provided duration must be positive.",
            {"duration": duration},
        )

    total_duration = duration or np.sum(pwc.durations)

    return graph.discretize_stf(
        stf=graph.convolve_pwc(
            pwc=pwc, kernel=graph.sinc_convolution_kernel(cutoff_frequency)
        ),
        duration=total_duration,
        segment_count=segment_count,
        name=name,
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

# pylint: disable=unused-argument, missing-function-docstring
"""
Deprecated functions.
"""



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Namespaces for the toolkit functions, nodes, and classes.

These namespaces are categorized by physical system.
"""

from dataclasses import dataclass
from enum import Enum

TOOLKIT_MAIN_DOC = """
The Boulder Opal Toolkits provide convenience nodes, classes, and functions that can
simplify developing and deploying workflows in Boulder Opal.

The toolkits are built on top of the existing Boulder Opal functions_ and graph operations_, and
are designed for a particular physical system, or system-agnostic tasks. For example,
the superconducting toolkit contains functionalities to simulate and optimize superconducting
qubit systems, and the signal library provides various forms of commonly used control signals.

Each toolkit provides convenience functions, which can be accessed through its corresponding
namespace from the :py:obj:`~qctrl.Qctrl` object. For example, all functions in the utility toolkit
live in the namespace of `utils` of the :py:obj:`~qctrl.Qctrl` object. For ease of use,
some toolkits also define classes for creating abstractions for the physical system.
For example, the `Cavity` class in the superconducting toolkit. These classes also live
in the corresponding `superconducting` namespace of the :py:obj:`~qctrl.Qctrl` object.

Finally, toolkits may also host convenience graph nodes, which can be used together with other
nodes_ for defining a general computation graph for your task. These operations
live in the corresponding namespace of a :py:obj:`~qctrl.graphs.Graph` object.
For example, you can access the nodes for defining signals from the `signals` namespace of
a :py:obj:`~qctrl.graphs.Graph` object.

For more context on the role of toolkits and their usage in Boulder Opal, see the
`Boulder Opal Toolkits <https://docs.q-ctrl.com/boulder-opal/topics/boulder-opal-toolkits>`_ topic.


.. _functions: https://docs.q-ctrl.com/boulder-opal/references/qctrl/Functions.html
.. _operations: https://docs.q-ctrl.com/boulder-opal/references/qctrl/Graphs.html
.. _nodes: https://docs.q-ctrl.com/boulder-opal/references/qctrl/Graphs.html

Following is a list of toolkits in Boulder Opal:
"""

_CLOSED_LOOP_DOC = """
Toolkit for closed-loop optimizations.

For a quick introduction, see the `Find optimal pulses with automated optimization
<https://docs.q-ctrl.com/boulder-opal/tutorials/find-optimal-pulses-with-automated-optimization>`_
tutorial and the `How to automate closed-loop hardware optimization
<https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-automate-closed-loop-hardware-optimization>`_
user guide.
"""

_IONS_DOC = """
Toolkit for trapped ion systems.
"""

_SIGNALS_DOC = """
Toolkit for signal library.

For a quick introduction, see the topic `Libraries of signals for Boulder Opal
<https://docs.q-ctrl.com/boulder-opal/topics/libraries-of-signals-for-boulder-opal>`_.
"""

_SUPERCONDUCTING_DOC = """
Toolkit for superconducting qubits.

For a quick introduction, see the `Simulate and optimize dynamics with the superconducting systems
toolkit <https://docs.q-ctrl.com/boulder-opal/tutorials/simulate-and-optimize-dynamics-
with-the-superconducting-systems-toolkit>`_ tutorial.
"""

_UTILS_DOC = """
Toolkit for system-agnostic functionality.
"""


@dataclass
class _NamespaceItem:
    """
    Store basic information about a namespace.
    """

    name: str
    doc: str
    title: str


class Namespace(Enum):
    """
    An enumeration of namespaces defined by physical systems.

    The `UTILS` namespace holds system-independent functionality.
    The `SIGNALS` namespace holds system-agnostic signals.
    """

    CLOSED_LOOP = _NamespaceItem(
        "closed_loop", _CLOSED_LOOP_DOC, "Closed-loop optimization"
    )
    IONS = _NamespaceItem("ions", _IONS_DOC, "Trapped ion systems")
    SIGNALS = _NamespaceItem("signals", _SIGNALS_DOC, "Signal library")
    SUPERCONDUCTING = _NamespaceItem(
        "superconducting", _SUPERCONDUCTING_DOC, "Superconducting systems"
    )
    UTILS = _NamespaceItem("utils", _UTILS_DOC, "General utilities")

    def get_name(self):
        """
        Get the name of the namespace.
        """
        return self.value.name

    def get_doc(self):
        """
        Get the doc of the namespace.
        """
        return self.value.doc

    def get_title(self):
        """
        Get the title of the doc page for toolkit.
        """
        if self.value.title is None:
            return f"{self.value.name.capitalize()} toolkit"

        return self.value.title



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""
Utilities for the closed-loop optimizer.
"""

from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Optional,
)

import numpy as np
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_integer,
    check_argument_non_negative_scalar,
)

from boulderopaltoolkits.namespace import Namespace
from boulderopaltoolkits.toolkit_utils import expose


@expose(Namespace.CLOSED_LOOP)
class Optimizer(ABC):
    """
    Abstract class for optimizers used in closed-loop control.

    To create an optimizer, use one of the concrete classes in the closed-loop optimization
    toolkit. For example to create a cross entropy optimizer with an elite fraction of 0.1:
    `optimizer=qctrl.closed_loop.CrossEntropy(elite_fraction=0.1)`.

    See Also
    --------
    :class:`.closed_loop.Cmaes` :
        Class describing the covariance matrix adaptation evolution strategy (CMA-ES) optimizer.
    :class:`.closed_loop.CrossEntropy` :
        Class describing the cross entropy optimizer.
    :class:`.closed_loop.GaussianProcess` :
        Class describing the Gaussian-process-based optimizer.
    :class:`.closed_loop.NeuralNetwork` :
        Class describing the neural-network-based optimizer.
    :class:`.closed_loop.SimulatedAnnealing` :
        Class describing the simulated annealing optimizer.
    """

    method_name: str

    @abstractmethod
    def create_optimizer(self, qctrl, bounds):
        """
        Return the initialized optimizer.

        Parameters
        ----------
        qctrl : qctrl.Qctrl
            Boulder Opal session object.
        bounds : np.ndarray
            The per-parameter bounds on the test points.
            The bounds must be a NumPy array of shape ``(parameter count, 2)`` where the
            trailing axis are the bounds for each parameter (with the lower bound first, followed
            by the upper bound).
        """
        raise NotImplementedError


@expose(Namespace.CLOSED_LOOP)
@dataclass
class Cmaes(Optimizer):
    r"""
    The covariance matrix adaptation evolution strategy (CMA-ES) optimizer.

    Parameters
    ----------
    initial_mean : np.ndarray, optional
        The array of the initial means of the parameters for the multivariate normal
        distribution.
        Defaults to None, in which case a random value inside the bounds
        is used for each parameter. If set, the value must be within the
        corresponding bounds for each parameter.
    initial_step_size : float, optional
        The initial step size for the multivariate normal distribution from which new test
        points are sampled.
        Defaults to one.
    population_size : int, optional
        The population size of the test candidates. It is recommended to use a population size of
        at least :math:`P = 4 + \lfloor 3 \times \log N \rfloor`, where :math:`N` is the number of
        optimizable parameters, :math:`\log` is the natural logarithm, and :math:`\lfloor x \rfloor`
        is the floor function.
        Defaults to :math:`P`.
    seed : int, optional
        Seed for the random number generator.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The CMA-ES optimizer uses a multivariate normal distribution to generate new test points.
    From an `initial_mean` and `initial_step_size`, this distribution is continually
    updated using an evolutionary strategy, with each update depending on the previous values and
    the current set of results.
    New test points are sampled from the distribution until convergence is reached.

    For more detail on CMA-ES see `CMA-ES <https://en.wikipedia.org/wiki/CMA-ES>`_ on Wikipedia.
    """

    initial_mean: Optional[np.ndarray] = None
    initial_step_size: Optional[float] = None
    population_size: Optional[int] = None
    seed: Optional[int] = None
    method_name = "CMA-ES"

    def __post_init__(self):
        if self.initial_mean is not None:
            check_argument(
                isinstance(self.initial_mean, np.ndarray)
                and self.initial_mean.ndim == 1,
                "The initial mean must be a 1D np.ndarray.",
                {"initial_mean": self.initial_mean},
            )
        if self.population_size is not None:
            check_argument(
                self.population_size > 0,
                "The population size must be at least one.",
                {"population_size": self.population_size},
            )

        if self.seed is not None:
            check_argument_integer(self.seed, "seed")

    def create_optimizer(self, qctrl: Any, bounds: np.ndarray):
        _check_bounds(bounds)

        if self.initial_mean is not None:
            check_argument(
                len(self.initial_mean) == len(bounds),
                "The initial mean and the bounds must have the same length.",
                {"initial_mean": self.initial_mean, "bounds": bounds},
            )
            for mean_val, bound in zip(self.initial_mean, bounds):
                check_argument(
                    bound[0] <= mean_val <= bound[1],
                    "The initial mean must be within the bounds.",
                    {"initial_mean": self.initial_mean, "bounds": bounds},
                )

        def convert_to_list(array):
            if array is None:
                return None
            return array.tolist()

        initializer = qctrl.types.closed_loop_optimization_step.CmaesInitializer(
            bounds=_convert_bounds(qctrl, bounds),
            initial_mean=convert_to_list(self.initial_mean),
            initial_step_size=self.initial_step_size,
            population_size=self.population_size,
            seed=self.seed,
        )
        return qctrl.types.closed_loop_optimization_step.Optimizer(
            cmaes_initializer=initializer
        )


@expose(Namespace.CLOSED_LOOP)
@dataclass
class CrossEntropy(Optimizer):
    r"""
    The cross-entropy optimizer.

    Parameters
    ----------
    elite_fraction : float
        The top fraction of test points that the algorithm uses to generate the next distribution.
        Must be between 0 and 1.
    seed : int, optional
        Seed for the random number generator.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The cross-entropy optimizer first fits a normal distribution to an `elite_fraction` of the given
    test points with the lowest cost function value. This normal distribution is then sampled to
    produce a new set of test points. Once all test points have converged, the optimization has
    converged and will no longer explore.

    Note that this optimizer requires at least
    :math:`\lceil 2 / \mathrm{elite}\_\mathrm{fraction} \rceil` results each step,
    to have at least 2 surviving elite points, and will throw an error if you don't provide enough
    `initial_test_parameters`.

    For best results, sample the initial test points from parameter space,
    varying each axis with each point. This optimizer will reject optimizations
    where all test points in the initial step are identical.

    For more detail on the cross-entropy optimizer see
    `cross-entropy method <https://en.wikipedia.org/wiki/Cross-entropy_method>`_ on Wikipedia.
    """

    elite_fraction: float
    seed: Optional[int] = None
    method_name = "cross entropy"

    def __post_init__(self):
        check_argument(
            0.0 < self.elite_fraction < 1.0,
            "The elite fraction must be between 0 and 1.",
            {"elite_fraction": self.elite_fraction},
        )

        if self.seed is not None:
            check_argument_integer(self.seed, "seed")

    def create_optimizer(self, qctrl: Any, bounds: np.ndarray):
        _check_bounds(bounds)

        initializer = qctrl.types.closed_loop_optimization_step.CrossEntropyInitializer(
            bounds=_convert_bounds(qctrl, bounds),
            elite_fraction=self.elite_fraction,
            seed=self.seed,
        )
        return qctrl.types.closed_loop_optimization_step.Optimizer(
            cross_entropy_initializer=initializer
        )


@expose(Namespace.CLOSED_LOOP)
@dataclass
class GaussianProcess(Optimizer):
    r"""
    The Gaussian process optimizer.

    Parameters
    ----------
    length_scale_bounds : np.ndarray, optional
        The per-parameter length scale bounds on the test points.
        The bounds must be a NumPy array of shape ``(parameter count, 2)`` where the trailing
        axis are the bounds for each parameter (with the lower bound first, followed by the upper
        bound).
        If not specified,  :py:obj:`~closed_loop.optimize` will pick a value derived from the
        `bounds` by picking orders of magnitudes below/above the sidelength for each box axis.
    seed : int, optional
        Seed for the random number generator.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The Gaussian process is defined by the kernel

    .. math::
        k({\mathbf x}_j, {\mathbf x}_k)
            = \exp \left(-\frac{1}{2} ( {\mathbf x}_j - {\mathbf x}_k )^\top
                \Sigma^{-2} ( {\mathbf x}_j - {\mathbf x}_k )\right) ,

    where :math:`{\mathbf x}_j` is an :math:`n`-dimensional vector representing the
    :math:`j`-th test point, :math:`\Sigma= {\rm diag}(l_1, \cdots, l_n)`
    is an :math:`n \times n` diagonal matrix, and :math:`\{ l_j \}` are the length scales.
    The length scales are tuned while training the model, within the bounds set by the
    `length_scale_bounds` parameter. Roughly speaking, the amount a parameter needs to change
    to impact the optimization cost should lie within the length scale bounds.

    It's recommended to provide non-zero `cost_uncertainty` to :py:obj:`~closed_loop.optimize`
    when using this optimizer, otherwise you might encounter a numerical error when the optimizer
    tries to fit the kernel with your input data. If the error persists, try increasing the
    `cost_uncertainty` value or decreasing the minimum length scale bound. However, such numerical
    error is also an indication that your data might not be suitable to be modelled by a
    Gaussian process, and in that case, consider using a different closed-loop optimizer.

    For more detail on Gaussian processes see
    `Gaussian process <https://en.wikipedia.org/wiki/Gaussian_process>`_ on Wikipedia.
    """

    length_scale_bounds: Optional[np.ndarray] = None
    seed: Optional[int] = None
    method_name = "Gaussian process"

    def __post_init__(self):
        if self.length_scale_bounds is not None:
            _check_bounds(self.length_scale_bounds, "length scale bounds")

        if self.seed is not None:
            check_argument_integer(self.seed, "seed")

    def create_optimizer(self, qctrl: Any, bounds: np.ndarray):
        _check_bounds(bounds)

        if self.length_scale_bounds is not None:
            check_argument(
                len(self.length_scale_bounds) == len(bounds),
                "The length scale bounds and the bounds must have the same length.",
                {"length_scale_bounds": self.length_scale_bounds, "bounds": bounds},
            )

        initializer = (
            qctrl.types.closed_loop_optimization_step.GaussianProcessInitializer(
                bounds=_convert_bounds(qctrl, bounds),
                length_scale_bounds=_convert_bounds(qctrl, self.length_scale_bounds),
                seed=self.seed,
            )
        )
        return qctrl.types.closed_loop_optimization_step.Optimizer(
            gaussian_process_initializer=initializer
        )


@expose(Namespace.CLOSED_LOOP)
@dataclass
class NeuralNetwork(Optimizer):
    """
    The neural network optimizer.

    Parameters
    ----------
    seed : int, optional
        Seed for the random number generator.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    The neural network optimizer builds and trains a neural network to fit the cost landscape with
    the data it receives. Then a set of test points are returned, which minimize the neural
    network's fitted cost landscape. A gradient based optimizer is used to minimize this landscape,
    with the points starting from different random initial values.

    This method is recommended when you can provide a large amount of data about your system.

    The network architecture used by this optimizer is chosen for its good performance on a variety
    of quantum control tasks.

    For best results, you should pass an array of `initial_test_parameters` evenly sampled
    over the whole parameter space.
    """

    seed: Optional[int] = None
    method_name = "neural network"

    def __post_init__(self):
        if self.seed is not None:
            check_argument_integer(self.seed, "seed")

    def create_optimizer(self, qctrl: Any, bounds: np.ndarray):
        _check_bounds(bounds)

        initializer = (
            qctrl.types.closed_loop_optimization_step.NeuralNetworkInitializer(
                bounds=_convert_bounds(qctrl, bounds), seed=self.seed
            )
        )
        return qctrl.types.closed_loop_optimization_step.Optimizer(
            neural_network_initializer=initializer
        )


@expose(Namespace.CLOSED_LOOP)
@dataclass
class SimulatedAnnealing(Optimizer):
    r"""
    The simulated annealing optimizer.

    Parameters
    ----------
    temperatures : np.ndarray
        The array of initial per-parameter annealing temperatures :math:`T_0` used to generate
        new test points.
        Higher temperatures correspond to higher exploration.
        The per-parameter adjustments from the current test point are sampled from Cauchy
        distributions with scales given by temperatures.
        The temperatures are currently implemented to decay such that each temperature
        at the k-th step is set according to :math:`T_k=\frac{T_0}{1+k}`.
        All temperatures must be positive.
    temperature_cost : float
        The parameter for controlling the optimizer’s greediness.
        A high cost temperature allows the optimizer to explore test points which may not
        immediately improve the cost. A higher level of exploration can be helpful for
        more difficult optimization problems. The cost temperature is set to decay
        according to the same schedule as the temperatures.
        Must be positive.
    seed : int, optional
        Seed for the random number generator.
        Use this option to generate deterministic results from the optimizer.

    Notes
    -----
    This simulated annealing performs a temperature based random walk within the parameter `bounds`.
    The new test points are sampled from a distribution whose variance is given by the current
    `temperatures`. Higher `temperatures` translate to high levels of exploration, which can be
    useful for non-convex optimization. The `temperature_cost` parameter can be set independently of
    the `temperatures`, and controls the overall greediness of each update. A high
    `temperature_cost` allows the optimizer to accept updates which do not immediately improve the
    cost. Both `temperatures` and the `temperature_cost` automatically decay between iterations.

    For more information on this method see
    `simulated annealing <https://en.wikipedia.org/wiki/Simulated_annealing>`_ on Wikipedia.
    """

    temperatures: np.ndarray
    temperature_cost: float
    seed: Optional[int] = None
    method_name = "simulated annealing"

    def __post_init__(self):
        check_argument(
            all(self.temperatures > 0),
            "All parameter temperatures must be positive.",
            {"temperatures": self.temperatures},
        )

        check_argument(
            self.temperature_cost > 0,
            "The cost temperature must be positive.",
            {"temperature_cost": self.temperature_cost},
        )

        if self.seed is not None:
            check_argument_integer(self.seed, "seed")

    def create_optimizer(self, qctrl: Any, bounds: np.ndarray):
        _check_bounds(bounds)

        check_argument(
            len(self.temperatures) == len(bounds),
            "The temperatures and the bounds must have the same length.",
            {"temperatures": self.temperatures, "bounds": bounds},
        )

        initializer = (
            qctrl.types.closed_loop_optimization_step.SimulatedAnnealingInitializer(
                bounds=_convert_bounds(qctrl, bounds),
                temperatures=list(self.temperatures),
                temperature_cost=self.temperature_cost,
                seed=self.seed,
            )
        )
        return qctrl.types.closed_loop_optimization_step.Optimizer(
            simulated_annealing_initializer=initializer
        )


def _convert_bounds(qctrl, bounds):
    """
    Convert bounds to box constraints, if the bounds are passed as a NumPy array.
    If the bounds are not a NumPy array None is returned.
    """
    if isinstance(bounds, np.ndarray):
        return [
            qctrl.types.closed_loop_optimization_step.BoxConstraint(
                lower_bound=float(bound[0]), upper_bound=float(bound[1])
            )
            for bound in bounds
        ]
    return None


def _check_bounds(bounds: np.ndarray, name: str = "bounds"):
    """Check that the bounds are well-defined."""
    check_argument(
        isinstance(bounds, np.ndarray) and bounds.ndim == 2 and bounds.shape[1] == 2,
        f"The {name} must be a 2D np.ndarray with two components in the second axis.",
        {name: bounds},
    )
    check_argument(
        all(bounds[:, 1] > bounds[:, 0]),
        "The upper bound (second component) must be greater than the lower bound "
        f"(first component) for each element in the {name} array.",
        {name: bounds},
    )


@expose(Namespace.CLOSED_LOOP)
def optimize(
    qctrl: Any,
    cost_function: Callable,
    initial_test_parameters: np.ndarray,
    optimizer: Optimizer,
    bounds: np.ndarray,
    cost_uncertainty: Optional[float] = None,
    target_cost: Optional[float] = None,
    max_iteration_count: int = 100,
    callback: Optional[Callable] = None,
    verbose: bool = True,
) -> dict:
    """
    Run a closed-loop optimization to find a minimum of the given cost function.

    This is an iterative process, where the optimizer generates and tests a set of points.
    After several iterations the distribution of generated test points should converge
    to low values of the cost function. You can use this approach when your system is too
    complicated to model, or the computation of gradient is expensive or impossible.

    Parameters
    ----------
    qctrl : qctrl.Qctrl
        Boulder Opal session object.
    cost_function : Callable
        A function that takes the parameters as an argument and returns an array of costs values.
        The function should take a NumPy array of input parameters with shape
        ``(test point count, parameter count)`` and return the costs in a 1D array of length test
        point count.
    initial_test_parameters : np.ndarray
        The initial values of the parameters to use in the optimization.
        A 2D NumPy array of shape ``(test point count, parameter count)``.
    optimizer : ~closed_loop.Optimizer
        The optimizer to be used in the minimization of the cost function.
    bounds : np.ndarray
        The per-parameter bounds on the test points.
        The bounds must be a NumPy array of shape ``(parameter count, 2)`` where the trailing
        axis are the bounds for each parameter (with the lower bound first, followed by the upper
        bound).
    cost_uncertainty : float, optional
        The standard deviation in the value of the cost.
        Must be non-negative.
        Defaults to 0.
    target_cost : float, optional
        The target cost.
        If passed, the optimization will halt if the best cost is below the given value.
    max_iteration_count : int, optional
        The maximum number of iterations.
        Defaults to 100.
    callback : Callable, optional
        A function that takes in the current set of parameters, a 2D NumPy array of shape
        ``(test point count, parameter count)``, and returns a bool.
        The function is evaluated once during each iteration with the
        current parameters. If it returns True, the optimization is halted.
    verbose : bool, optional
        Whether to print out information about the optimization cycle.
        Defaults to True.

    Returns
    -------
    dict
        A dictionary containing the results of the optimization, namely, the best parameters
        `best_parameters`, their associated cost `best_cost`, and
        the history of best cost values `best_cost_history`.

    Notes
    -----
    If the optimization loop is halted via a `KeyboardInterrupt` then the function returns
    the best results obtained in the optimization thus far.
    """

    def verbose_print(message):
        if verbose:
            print(message)

    check_argument(
        isinstance(initial_test_parameters, np.ndarray)
        and np.ndim(initial_test_parameters) == 2,
        "The initial test parameters must be a 2D np.ndarray.",
        {"initial_test_parameters": initial_test_parameters},
    )

    check_argument(
        np.all(
            (initial_test_parameters >= bounds[:, 0])
            & (initial_test_parameters <= bounds[:, 1])
        ),
        "The initial test parameters must be within the bounds.",
        {"initial_test_parameters": initial_test_parameters, "bounds": bounds},
    )

    if cost_uncertainty is not None:
        check_argument_non_negative_scalar(cost_uncertainty, "cost uncertainty")

    verbose_print(
        f"""Running closed loop optimization
----------------------------------------
  Optimizer            : {optimizer.method_name}
  Number of test points: {initial_test_parameters.shape[0]}
  Number of parameters : {initial_test_parameters.shape[1]}
----------------------------------------
"""
    )

    closed_loop_optimizer = optimizer.create_optimizer(qctrl, bounds)

    test_parameters = initial_test_parameters

    # Obtain initial costs.
    verbose_print("Calling cost function…")
    costs = cost_function(test_parameters)

    check_argument(
        len(np.asarray(costs).shape) == 1
        and len(costs) == len(initial_test_parameters),
        "The cost function must return a 1D array of the length test point count.",
        {"initial_test_parameters": initial_test_parameters},
        extras={"cost_function(test_parameters)": costs},
    )

    best_cost_overall, best_parameters_overall = min(
        zip(costs, test_parameters), key=lambda params: params[0]
    )
    verbose_print(f"  Initial best cost: {best_cost_overall:.3e}")

    # Store the cost history.
    best_cost_history = [best_cost_overall]

    # Run the optimization loop until a halting condition is met.
    try:
        for iteration_count in range(max_iteration_count):
            # Organize the costs into the proper input format.
            results = [
                qctrl.types.closed_loop_optimization_step.CostFunctionResult(
                    parameters=parameters, cost=cost, cost_uncertainty=cost_uncertainty
                )
                for parameters, cost in zip(test_parameters, costs)
            ]

            # Call the automated closed-loop optimizer and obtain the next set of test points.
            verbose_print("\nRunning optimizer…")
            optimization_result = (
                qctrl.functions.calculate_closed_loop_optimization_step(
                    optimizer=closed_loop_optimizer,
                    results=results,
                    test_point_count=test_parameters.shape[0],
                )
            )

            # Retrieve the optimizer state and create a new optimizer object.
            closed_loop_optimizer = qctrl.types.closed_loop_optimization_step.Optimizer(
                state=optimization_result.state
            )

            # Organize the data returned by the automated closed-loop optimizer.
            test_parameters = np.array(
                [
                    test_point.parameters
                    for test_point in optimization_result.test_points
                ]
            )

            # Obtain costs.
            verbose_print("Calling cost function…")
            costs = cost_function(test_parameters)

            # Record the best results after this round.
            best_cost, best_parameters = min(
                zip(costs, test_parameters), key=lambda params: params[0]
            )

            # Compare last best results with best result overall.
            if best_cost < best_cost_overall:
                best_cost_overall = best_cost
                best_parameters_overall = best_parameters

            # Print the current best cost.
            verbose_print(
                f"  Best cost after {iteration_count+1} iterations: {best_cost_overall:.3e}"
            )

            # Store the current best cost.
            best_cost_history.append(best_cost_overall)

            if callback is not None:
                if callback(test_parameters):
                    verbose_print(
                        "\nCallback condition satisfied. Stopping the optimization."
                    )
                    break

            # Check if desired threshold has been achieved.
            if target_cost is not None:
                if best_cost_overall < target_cost:
                    verbose_print("\nTarget cost reached. Stopping the optimization.")
                    break
        else:
            verbose_print(
                "\nMaximum iteration count reached. Stopping the optimization."
            )

    except KeyboardInterrupt:
        # Exit the loop and return the current best results if a KeyboardInterrupt is caught.
        print("\nCalculation interrupted. Stopping the optimization.")

    return {
        "best_cost": best_cost_overall,
        "best_parameters": best_parameters_overall,
        "best_cost_history": best_cost_history,
    }



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Utility functions for exposing and binding toolkits.
"""
from __future__ import annotations

import inspect
from enum import Enum
from functools import (
    partial,
    wraps,
)
from typing import (
    Any,
    Callable,
    Optional,
)

import forge
import numpy as np
from numpydoc.docscrape import NumpyDocString
from qctrlcommons.exceptions import QctrlException
from qctrlcommons.node.node_data import NodeData
from qctrlcommons.preconditions import check_argument

from boulderopaltoolkits.namespace import Namespace

# parameters like qctrl/graph in the toolkit functions are just placeholder for the actual object
# need to remove them from the doc/signature before binding in python
_EXCLUDE_PARAMETERS = {"qctrl", "graph"}

TOOLKIT_ATTR = "_toolkit_attr"


def expose(*namespaces: Namespace) -> Any:
    """
    Define a decorator that inserts a label of the namespaces for later binding to the
    wrapped object.

    The decorator sets a new attribute defined by TOOLKIT_ATTR to the object, which is a tuple
    of strings defining the namespaces. You can apply this decorator to the toolkit function/class,
    and they will be bound to the corresponding namespaces when initializing the Qctrl/Graph object.
    It also sets a new attribute to the object to allow tracking through the python package.

    For example, to expose a function to one namespace ::

        @expose(Namespace.SUPERCONDUCTING)
        def some_func()

    To expose a function to multiple namespaces, as it maybe be reused for
    different physical systems ::

        @expose(Namespace.SUPERCONDUCTING, Namespace.IONS)
        def some_general_func()
    """
    return partial(_expose, attr=namespaces)


def _expose(obj, attr):
    setattr(obj, TOOLKIT_ATTR, attr)
    namespaces = "_".join([f"{i.name.lower()}" for i in attr])
    method_origin = f"boulder-opal-toolkits.{namespaces}.{getattr(obj, '__name__')}"
    setattr(obj, "method_origin", method_origin)
    return obj


def forge_toolkit(toolkit: Any, obj: Any) -> Any:
    """
    Forge the toolkit object (must be a function or class) for binding.

    Note that when toolkit is a class, we just return it, since it should
    not accept the to-be-bound object, to avoid the cyclic access.

    Parameters
    ----------
    toolkit : Any
        A toolkit function or class to be forged.
    obj : Any
        The object to which the forged toolkit is bound.

    Returns
    -------
    Any
        A forged class/function.

    Raises
    ------
    QctrlException
        If toolkit has an unsupported type.
    """

    if inspect.isfunction(toolkit):
        return _forge_func(toolkit, obj)
    if inspect.isclass(toolkit):
        return toolkit
    raise QctrlException(
        f"Toolkit must either be a function or a class, but got {type(toolkit)}."
    )


def _forge_func(func: Callable, obj: Any) -> Callable:
    """
    Forge toolkit function.
    """

    def forged(*arg, **kwargs):
        if hasattr(obj, "method_origin"):
            obj.method_origin = getattr(func, "method_origin", None)
            result = func(obj, *arg, **kwargs)
            obj.method_origin = None
        else:
            result = func(obj, *arg, **kwargs)

        return result

    _args, _kwargs, _return_type, bind_obj = _get_args_kwargs_return_type(func)

    if not bind_obj:
        return func

    assert func.__doc__ is not None, f"{func} has no doc."
    forged.__doc__ = _clean_doc(func.__doc__)
    forged.__name__ = func.__name__
    sig = forge.sign(*_args, **_kwargs)
    forged = sig(forged)
    forged = forge.returns(_return_type)(forged)

    return forged


def _get_args_kwargs_return_type(func):
    """
    Get the information about the function signature.

    Returns a tuple, where the first element is list of args,
    the second one is a dictionary for kwargs,
    the third one is the type of the returned value,
    and the last one is a boolean to exclude parameters.
    This will be used to forge a new function for binding.
    """
    _args = []
    _kwargs = {}
    bind_obj = False
    _func_inspected = inspect.signature(func)
    for arg_name, _arg in _func_inspected.parameters.items():
        # note that some convenience function might not be fetchable
        if arg_name == "name":
            _kwargs = {"name": forge.kwarg("name", type=Optional[str], default=None)}
        elif arg_name in _EXCLUDE_PARAMETERS:
            bind_obj = True
        else:
            _args.append(
                forge.arg(
                    arg_name, type=_arg.annotation, default=_get_default_arg_val(_arg)
                )
            )

    return _args, _kwargs, _func_inspected.return_annotation, bind_obj


def _get_default_arg_val(arg: inspect.Parameter) -> Any:
    """
    Post process the default value for a given argument:
        - If the default value is an Enum, use the value of the Enum.
        - Otherwise, do nothing.
    """
    default_val = arg.default
    if isinstance(default_val, Enum):
        return default_val.value
    return default_val


def _clean_doc(doc: str) -> str:
    """
    Removes unnecessary part from the doc.
    """
    doc_obj = NumpyDocString(doc)

    doc_obj["Parameters"] = [
        item for item in doc_obj["Parameters"] if item.name not in _EXCLUDE_PARAMETERS
    ]

    return str(doc_obj)


def no_graph_parameters(alternatives: Optional[str]) -> Callable:
    """
    Check that none of the parameters passed to the function are graph objects.

    Use this as a decorator for non-graph functions, to make sure that no
    graph object types are being passed as parameters by accident.

    Parameters
    ----------
    alternatives : str or None
        Text listing alternative functions that do accept graphs and that the
        user might be intending to use instead. If it's None, nothing is
        suggested to the user.

    Returns
    -------
    Callable
        A decorator that adds checks to the beginning of the function to see
        if none of the inputs are graph types, and raises an exception that
        suggests the `alternatives`.

    Examples
    --------
    Create a addition function that only accepts non-Graph objects as input,
    and that suggests `graph.add` as an alternative.

    >>> @no_graph_parameters(alternatives="`graph.add`")
    ... def numpy_add(x, y):
    ...     return x + y
    >>> numpy_add(1, graph.tensor(2))
    qctrlcommons.exceptions.QctrlArgumentsValueError: The parameters of
    `numpy_add` cannot be graph objects. Perhaps you intended to use `graph.add`?
    y=<Tensor: name="tensor_#1", operation_name="tensor", shape=()>
    type(y)='Tensor'
    """
    advice = ""
    if alternatives is not None:
        advice = f" Perhaps you intended to use {alternatives}?"

    def _no_graph_parameters(function: Callable):
        arg_names = list(inspect.signature(function).parameters)

        error_message = (
            f"The parameters of `{function.__name__}` cannot be graph objects." + advice
        )

        @wraps(function)
        def _function(*args, **kwargs):
            for arg_name, arg in zip(arg_names, args):
                check_argument(
                    not isinstance(arg, NodeData),
                    error_message,
                    {arg_name: arg},
                    extras={f"type({arg_name})": type(arg).__name__},
                )

            for key, kwarg in kwargs.items():
                check_argument(
                    not isinstance(kwarg, NodeData),
                    error_message,
                    {key: kwarg},
                    extras={f"type({key})": type(kwarg).__name__},
                )
            return function(*args, **kwargs)

        return _function

    return _no_graph_parameters


def extract_graph_output(item):
    """
    Convert a non-batched item from a qctrl.types.graph/optimization.Result.output
    dictionary into a NumPy array.

    If the item is a dictionary (corresponds to a tensor), then its "value" is returned.

    If the item is a list of dictionaries (corresponds to a Pwc), then the "value"
    elements of its inner dictionaries are returned in a NumPy array.

    Parameters
    ----------
    item : dict or list
        The item of a Result.output dictionary representing a tensor or a PWC node.

    Returns
    -------
    np.ndarray or number
        The value of the tensor node or the values of the PWC node.
    """
    if isinstance(item, dict):
        return item["value"]
    if isinstance(item, list):
        assert all(
            isinstance(elem, dict) for elem in item
        ), "Batches are not supported."
        return np.array([seg["value"] for seg in item])
    assert False, "Unknown output format."



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
STF signal library nodes.
"""

from __future__ import annotations

from typing import Optional

import numpy as np
from qctrlcommons.graph import Graph
from qctrlcommons.node.node_data import (
    Stf,
    Tensor,
)
from qctrlcommons.preconditions import check_argument

from boulderopaltoolkits.namespace import Namespace
from boulderopaltoolkits.signals.signal_utils import validate_optimizable_parameter
from boulderopaltoolkits.toolkit_utils import expose


@expose(Namespace.SIGNALS)
def sech_pulse_stf(
    graph: Graph,
    amplitude: float | complex | Tensor,
    width: float | Tensor,
    center_time: float | Tensor,
) -> Stf:
    r"""
    Create an `Stf` representing a hyperbolic secant pulse.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    amplitude : float or complex or Tensor
        The amplitude of the pulse, :math:`A`.
        It must either be a scalar or contain a single element.
    width : float or Tensor
        The characteristic time for the hyperbolic secant pulse, :math:`t_\mathrm{pulse}`.
        It must either be a scalar or contain a single element.
    center_time : float or Tensor
        The time at which the pulse peaks, :math:`t_\mathrm{peak}`.
        It must either be a scalar or contain a single element.

    Returns
    -------
    Stf
        The sampleable hyperbolic secant pulse.

    See Also
    --------
    :func:`.signals.gaussian_pulse_stf` : Create an `Stf` representing a Gaussian pulse.
    :func:`.signals.sech_pulse` :
        Function to create a `Signal` object representing a hyperbolic secant pulse.
    :func:`.signals.sech_pulse_pwc` : Corresponding operation with `Pwc` output.

    Notes
    -----
    The hyperbolic secant pulse is defined as

        .. math:: \mathop{\mathrm{Sech}}(t)
            = \frac{A}{\cosh\left((t - t_\mathrm{peak}) / t_\mathrm{pulse} \right)} .

    The full width at half maximum of the pulse is about :math:`2.634 t_\mathrm{pulse}`.

    Examples
    --------
    Define a sampleable sech pulse.

    >>> sech = graph.signals.sech_pulse_stf(
    ...     amplitude=1.0, width=0.1, center_time=0.5
    ... )
    >>> sech
    <Stf: operation_name="truediv", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(stf=sech, sample_times=np.linspace(0, 1, 5), name="sech_samples")
    <Tensor: name="sech_samples", operation_name="sample_stf", shape=(5,)>
    >>> graph.discretize_stf(stf=sech, duration=1.2, segment_count=100, name="discretized_sech")
    <Pwc: name="discretized_sech", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["sech_samples", "discretized_sech"]
    ... )
    >>> result.output["sech_samples"]["value"]
    array([0.013, 0.163, 1.000, 0.163, 0.013])

    Define a sampleable sech pulse with optimizable parameters.

    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
    ... )
    >>> width = graph.optimization_variable(
    ...     count=1, lower_bound=0.1, upper_bound=0.5, name="width"
    ... )
    >>> center_time = graph.optimization_variable(
    ...     count=1, lower_bound=0.2, upper_bound=0.8, name="center_time"
    ... )
    >>> graph.signals.sech_pulse_stf(
    ...     amplitude=amplitude, width=width, center_time=center_time
    ... )
    <Stf: operation_name="truediv", value_shape=(), batch_shape=()>
    """

    amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
    width = validate_optimizable_parameter(graph, width, "width")
    center_time = validate_optimizable_parameter(graph, center_time, "center time")

    return amplitude / graph.cosh((graph.identity_stf() - center_time) / width)


@expose(Namespace.SIGNALS)
def gaussian_pulse_stf(
    graph: Graph,
    amplitude: float | complex | Tensor,
    width: float | Tensor,
    center_time: float | Tensor,
    drag: Optional[float | Tensor] = None,
) -> Stf:
    # pylint: disable=line-too-long
    r"""
    Create an `Stf` representing a Gaussian pulse.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    amplitude : float or complex or Tensor
        The amplitude of the Gaussian pulse, :math:`A`.
        It must either be a scalar or contain a single element.
    width : float or Tensor
        The standard deviation of the Gaussian pulse, :math:`\sigma`.
        It must either be a scalar or contain a single element.
    center_time : float or Tensor
        The center of the Gaussian pulse, :math:`t_0`.
        It must either be a scalar or contain a single element.
    drag : float or Tensor, optional
        The DRAG parameter, :math:`\beta`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to no DRAG correction.

    Returns
    -------
    Stf
        The sampleable Gaussian pulse.

    See Also
    --------
    :func:`.signals.gaussian_pulse_pwc` : Corresponding operation with `Pwc` output.
    :func:`.signals.sech_pulse_stf` : Create an `Stf` representing a hyperbolic secant pulse.

    Notes
    -----
    The Gaussian pulse is defined as

    .. math:: \mathop{\mathrm{Gaussian}}(t) =
        A \left(1-\frac{i\beta (t-t_0)}{\sigma^2}\right)
        \exp \left(- \frac{(t-t_0)^2}{2\sigma^2} \right) .

    Examples
    --------
    Define a sampleable Gaussian pulse.

    >>> gaussian = graph.signals.gaussian_pulse_stf(
    ...     amplitude=1.0, width=0.1, center_time=0.5
    ... )
    >>> gaussian
    <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(stf=gaussian, sample_times=np.linspace(0, 1, 5), name="gaussian_samples")
    <Tensor: name="gaussian_samples", operation_name="sample_stf", shape=(5,)>
    >>> graph.discretize_stf(
    ... gaussian, duration=1, segment_count=100, name="discretized_gaussian"
    ... )
    <Pwc: name="discretized_gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["gaussian_samples", "discretized_gaussian"]
    ... )
    >>> result.output["gaussian_samples"]["value"]
    array([3.727e-06, 4.394e-02, 1.000e+00, 4.394e-02, 3.727e-06])

    Define a sampleable Gaussian with a DRAG correction.

    >>> gaussian = graph.signals.gaussian_pulse_stf(
    ...     amplitude=1.0, width=0.1, center_time=0.5, drag=0.2
    ... )
    >>> gaussian
    <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(
    ...     stf=gaussian, sample_times=np.linspace(0, 1, 5), name="drag_gaussian_samples"
    ... )
    <Tensor: name="drag_gaussian_samples", operation_name="sample_stf", shape=(5,)>
    >>> graph.discretize_stf(
    ...     gaussian, duration=1, segment_count=100, name="discretized_drag_gaussian"
    ... )
    <Pwc: name="discretized_drag_gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["drag_gaussian_samples", "discretized_drag_gaussian"]
    ... )
    >>> result.output["drag_gaussian_samples"]["value"]
    array([
        3.727e-06 + 9.317e-06j,
        4.394e-02 + 5.492e-02j,
        1.000e+00 + 0.000e+00j,
        4.394e-02 - 5.492e-02j,
        3.727e-06 - 9.317e-06j,
    ])

    Define a sampleable Gaussian pulse with optimizable parameters.

    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
    ... )
    >>> width = graph.optimization_variable(
    ...     count=1, lower_bound=0.1, upper_bound=0.5, name="width"
    ... )
    >>> center_time = graph.optimization_variable(
    ...     count=1, lower_bound=0.2, upper_bound=0.8, name="center_time"
    ... )
    >>> drag = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=0.5, name="drag"
    ... )
    >>> graph.signals.gaussian_pulse_stf(
    ...     amplitude=amplitude, width=width, center_time=center_time, drag=drag
    ... )
    <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
    """
    # pylint: enable=line-too-long

    amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
    width = validate_optimizable_parameter(graph, width, "width")
    center_time = validate_optimizable_parameter(graph, center_time, "center time")

    if drag is not None:
        drag = validate_optimizable_parameter(graph, drag, "drag")
        correction = -(1j * drag / (width**2)) * (graph.identity_stf() - center_time)
        amplitude *= 1.0 + correction

    return amplitude * graph.exp(
        -((graph.identity_stf() - center_time) ** 2) / (2 * width**2)
    )


@expose(Namespace.SIGNALS)
def sinusoid_stf(
    graph: Graph,
    amplitude: float | complex | Tensor,
    angular_frequency: float | Tensor,
    phase: float | Tensor = 0.0,
) -> Stf:
    # pylint: disable=line-too-long
    r"""
    Create an `Stf` representing a sinusoidal oscillation.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    amplitude : float or complex or Tensor
        The amplitude of the oscillation, :math:`A`.
        It must either be a scalar or contain a single element.
    angular_frequency : float or Tensor
        The angular frequency of the oscillation, :math:`\omega`.
        It must either be a scalar or contain a single element.
    phase : float or Tensor, optional
        The phase of the oscillation, :math:`\phi`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to 0.

    Returns
    -------
    Stf
        The sampleable sinusoid.

    See Also
    --------
    :func:`.signals.hann_series_stf` : Create an `Stf` representing a sum of Hann window functions.
    :func:`.signals.sinusoid` :
        Function to create a `Signal` object representing a sinusoidal oscillation.
    :func:`.signals.sinusoid_pwc` : Corresponding operation with `Pwc` output.
    :func:`~qctrl.graphs.Graph.sin` : Calculate the element-wise sine of an object.

    Notes
    -----
    The sinusoid is defined as

    .. math:: \mathop{\mathrm{Sinusoid}}(t) = A \sin \left( \omega t + \phi \right) .

    Examples
    --------
    Define an STF oscillation.

    >>> oscillation = graph.signals.sinusoid_stf(
    ...     amplitude=1.0, angular_frequency=np.pi, phase=np.pi/2.0
    ... )
    >>> oscillation
    <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(stf=oscillation, sample_times=np.linspace(0, 1, 5), name="oscillation")
    <Tensor: name="oscillation", operation_name="sample_stf", shape=(5,)>
    >>> graph.discretize_stf(
    ...     oscillation, duration=10, segment_count=100, name="discretized_oscillation"
    ... )
    <Pwc: name="discretized_oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["oscillation", "discretized_oscillation"]
    ... )
    >>> result.output["oscillation"]["value"]
    array([ 1.000e+00,  7.071e-01,  1.225e-16, -7.071e-01, -1.000e+00])

    Define a sinusoid with optimizable parameters.

    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=4e3, name="amplitude"
    ... )
    >>> angular_frequency = graph.optimization_variable(
    ...     count=1, lower_bound=5e6, upper_bound=20e6, name="angular_frequency"
    ... )
    >>> phase = graph.optimization_variable(
    ...     count=1,
    ...     lower_bound=0,
    ...     upper_bound=2*np.pi,
    ...     is_lower_unbounded=True,
    ...     is_upper_unbounded=True,
    ...     name="phase",
    ... )
    >>> graph.signals.sinusoid_stf(
    ...     amplitude=amplitude, angular_frequency=angular_frequency, phase=phase
    ... )
    <Stf: operation_name="multiply", value_shape=(), batch_shape=()>
    """
    # pylint: enable=line-too-long

    amplitude = validate_optimizable_parameter(graph, amplitude, "amplitude")
    angular_frequency = validate_optimizable_parameter(
        graph, angular_frequency, "angular frequency"
    )
    phase = validate_optimizable_parameter(graph, phase, "phase")

    return amplitude * graph.sin(angular_frequency * graph.identity_stf() + phase)


@expose(Namespace.SIGNALS)
def hann_series_stf(
    graph: Graph,
    coefficients: np.ndarray | Tensor,
    end_time: float,
    start_time: float = 0.0,
) -> Stf:
    r"""
    Create an `Stf` representing a sum of Hann window functions.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    coefficients : np.ndarray or Tensor
        The coefficients for the different Hann window functions, :math:`c_n`.
        It must be a 1D array or Tensor.
    end_time : float
        The time at which the Hann series ends, :math:`t_\mathrm{end}`.
    start_time : float, optional
        The time at which the Hann series starts, :math:`t_\mathrm{start}`.
        Defaults to 0.

    Returns
    -------
    Stf
        The sampleable Hann window functions series.

    See Also
    --------
    :func:`.signals.hann_series` :
        Function to create a `Signal` object representing a sum of Hann window functions.
    :func:`.signals.hann_series_pwc` : Corresponding operation with `Pwc` output.
    :func:`.signals.sinusoid_stf` : Create an `Stf` representing a sinusoidal oscillation.

    Notes
    -----
    The series is defined as

    .. math:: \mathop{\mathrm{Hann}}(t)
        = \sum_{n=1}^N c_n \sin^2 \left(
            \frac{\pi n (t - t_\mathrm{start})}{t_\mathrm{end} - t_\mathrm{start}}
        \right) ,

    where :math:`N` is the number of coefficients.

    Note that the function values outside the :math:`(t_\mathrm{start}, t_\mathrm{end})` range
    will not be zero.

    Examples
    --------
    Define a simple sampleable Hann series.

    >>> hann = graph.signals.hann_series_stf(coefficients=np.array([0.5, 1, 0.25]), end_time=1.0)
    >>> hann
    <Stf: operation_name="stf_sum", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(stf=hann, sample_times=np.linspace(0, 1, 7), name="hann_samples")
    <Tensor: name="hann_samples", operation_name="hann_samples", shape=(7,)>
    >>> graph.discretize_stf(hann, duration=1, segment_count=100, name="discretized_hann")
    <Pwc: name="discretized_hann", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["hann_samples", "discretized_hann"]
    ... )
    >>> result.output["hann_samples"]["value"]
    array([0.000e+00, 1.125e+00, 1.125e+00, 7.500e-01, 1.125e+00, 1.125e+00, 5.159e-14])

    Define a sampleable Hann series with optimizable coefficients.

    >>> coefficients = graph.optimization_variable(
    ...     count=8, lower_bound=-3.5e6, upper_bound=3.5e6, name="coefficients"
    ... )
    >>> graph.signals.hann_series_stf(coefficients=coefficients, end_time=2.0e-6)
    <Stf: operation_name="stf_sum", value_shape=(), batch_shape=()>
    """

    check_argument(
        len(coefficients.shape) == 1,
        "The coefficients must be in a 1D array or Tensor.",
        {"coefficients": coefficients},
        extras={"coefficients.shape": coefficients.shape},
    )

    check_argument(
        end_time > start_time,
        "The end time must be greater than the start time.",
        {"start_time": start_time, "end_time": end_time},
    )

    # Define scaled times π (t - t_start) / (t_end - t_start).
    scaled_time = (graph.identity_stf() - start_time) * (
        np.pi / (end_time - start_time)
    )

    # Calculate function values.
    stfs = [
        coefficients[idx] * graph.sin((idx + 1) * scaled_time) ** 2
        for idx in range(coefficients.shape[0])
    ]
    return graph.stf_sum(stfs)


@expose(Namespace.SIGNALS)
def linear_ramp_stf(
    graph: Graph, slope: float | complex | Tensor, shift: float | complex | Tensor = 0.0
) -> Stf:
    # pylint:disable=line-too-long
    r"""
    Create an `Stf` representing a linear ramp.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    slope : float or complex or Tensor
        The slope of the ramp, :math:`a`.
        It must either be a scalar or contain a single element.
    shift : float or complex or Tensor, optional
        The value of the ramp at :math:`t = 0`, :math:`b`.
        It must either be a scalar or contain a single element.
        Defaults to 0.

    Returns
    -------
    Stf
        The sampleable linear ramp.

    See Also
    --------
    :func:`.signals.linear_ramp` : Function to create a `Signal` object representing a linear ramp.
    :func:`.signals.linear_ramp_pwc` : Corresponding operation with `Pwc` output.
    :func:`.signals.tanh_ramp_stf` : Create an `Stf` representing a hyperbolic tangent ramp.

    Notes
    -----
    The linear ramp is defined as

    .. math:: \mathop{\mathrm{Linear}}(t) = a t + b .

    Examples
    --------
    Define a linear STF ramp.

    >>> linear = graph.signals.linear_ramp_stf(slope=4.0, shift=-2.0)
    >>> linear
    <Stf: operation_name="add", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(stf=linear, sample_times=np.linspace(0, 1, 5), name="linear_ramp")
    <Tensor: name="linear_ramp", operation_name="sample_stf", shape=(5,)>
    >>> graph.discretize_stf(linear, duration=1, segment_count=100, name="discretized_linear")
    <Pwc: name="discretized_linear", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["linear_ramp", "discretized_linear"]
    ... )
    >>> result.output["linear_ramp"]["value"]
    array([-2., -1., 0., 1., 2.]

    Define a linear STF ramp with an optimizable slope and root.

    >>> slope = graph.optimization_variable(
    ...     count=1, lower_bound=-4, upper_bound=4, name="slope"
    ... )
    >>> root = graph.optimization_variable(
    ...     count=1, lower_bound=-4, upper_bound=4, name="slope"
    ... )
    >>> shift = - slope * root
    >>> graph.signals.linear_ramp_stf(slope=slope, shift=shift)
    <Stf: operation_name="add", value_shape=(), batch_shape=()>
    """
    # pylint:enable=line-too-long

    slope = validate_optimizable_parameter(graph, slope, "slope")
    shift = validate_optimizable_parameter(graph, shift, "shift")

    return slope * graph.identity_stf() + shift


@expose(Namespace.SIGNALS)
def tanh_ramp_stf(
    graph: Graph,
    center_time: float | Tensor,
    ramp_duration: float | Tensor,
    end_value: float | complex | Tensor,
    start_value: Optional[float | complex | Tensor] = None,
) -> Stf:
    r"""
    Create an `Stf` representing a hyperbolic tangent ramp.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    center_time : float or Tensor, optional
        The time at which the ramp has its greatest slope, :math:`t_0`.
        It must either be a scalar or contain a single element.
    ramp_duration : float or Tensor, optional
        The characteristic time for the hyperbolic tangent ramp, :math:`t_\mathrm{ramp}`.
        It must either be a scalar or contain a single element.
    end_value : float or complex or Tensor
        The asymptotic value of the ramp towards :math:`t \to +\infty`, :math:`a_+`.
        It must either be a scalar or contain a single element.
    start_value : float or complex or Tensor, optional
        The asymptotic value of the ramp towards :math:`t \to -\infty`, :math:`a_-`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to minus `end_value`.

    Returns
    -------
    Stf
        The sampleable hyperbolic tangent ramp.

    See Also
    --------
    :func:`.signals.linear_ramp_stf` : Create an `Stf` representing a linear ramp.
    :func:`.signals.tanh_ramp` :
        Function to create a `Signal` object representing a hyperbolic tangent ramp.
    :func:`.signals.tanh_ramp_pwc` : Corresponding operation with `Pwc` output.
    :func:`~qctrl.graphs.Graph.tanh` : Calculate the element-wise hyperbolic tangent of an object.

    Notes
    -----
    The hyperbolic tangent ramp is defined as

    .. math:: \mathop{\mathrm{Tanh}}(t)
        = \frac{a_+ + a_-}{2}
            + \frac{a_+ - a_-}{2} \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

    where the function's asymptotic values :math:`a_\pm` are defined by:

    .. math::  a_\pm := \lim_{t\to\pm\infty} \mathop{\mathrm{Tanh}}(t) ,

    and :math:`t_0` is related to :math:`t_\mathrm{ramp}` by:

    .. math::
        \left.\frac{{\rm d}\mathop{\mathrm{Tanh}}(t)}{{\rm d}t}\right|_{t=t_0}
            = \frac{ (a_+ - a_-)}{2 t_\mathrm{ramp}} .

    With the default value of `start_value` (:math:`a_-`),
    the ramp expression simplifies to

    .. math:: \mathop{\mathrm{Tanh}}(t)
        = A \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

    where :math:`A = a_+` is the end value (the start value is then :math:`-A`).

    Examples
    --------
    Define a simple sampleable hyperbolic tangent ramp.

    >>> tanh = graph.signals.tanh_ramp_stf(
    ...     center_time=0.4, ramp_duration=0.2, end_value=2, start_value=-1
    ... )
    >>> tanh
    <Stf: operation_name="add", value_shape=(), batch_shape=()>
    >>> graph.sample_stf(stf=tanh, sample_times=np.linspace(0, 1, 7), name="tanh_samples")
    <Tensor: name="tanh_samples", operation_name="sample_stf", shape=(7,)>
    >>> graph.discretize_stf(tanh, duration=1, segment_count=100, name="discretized_tanh")
    <Pwc: name="discretized_tanh", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["tanh_samples", "discretized_tanh"]
    ... )
    >>> result.output["tanh_samples"]["value"]
    array([-0.946, -0.735,  0.018,  1.193,  1.805,  1.961,  1.993])

    Define a hyperbolic tangent ramp with optimizable parameters.

    >>> center_time = graph.optimization_variable(
    ...     count=1, lower_bound=0.25e-6, upper_bound=0.75e-6, name="center_time"
    ... )
    >>> ramp_duration = graph.optimization_variable(
    ...     count=1, lower_bound=0.1e-6, upper_bound=0.3e-6, name="ramp_duration"
    ... )
    >>> end_value = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=3e6, name="end_value"
    ... )
    >>> graph.signals.tanh_ramp_stf(
    ...     center_time=center_time, ramp_duration=ramp_duration, end_value=end_value
    ... )
     <Stf: operation_name="add", value_shape=(), batch_shape=()>
    """

    if start_value is None:
        start_value = -end_value

    center_time = validate_optimizable_parameter(graph, center_time, "center time")
    ramp_duration = validate_optimizable_parameter(
        graph, ramp_duration, "ramp duration"
    )
    end_value = validate_optimizable_parameter(graph, end_value, "end value")
    start_value = validate_optimizable_parameter(graph, start_value, "start value")

    return start_value + 0.5 * (end_value - start_value) * (
        1 + graph.tanh((graph.identity_stf() - center_time) / ramp_duration)
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Signal library functions.
"""
from __future__ import annotations

from enum import Enum
from typing import (
    Callable,
    Optional,
)

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopaltoolkits.namespace import Namespace
from boulderopaltoolkits.toolkit_utils import (
    expose,
    no_graph_parameters,
)


@expose(Namespace.SIGNALS)
class Signal:
    """
    A class that contains information about a signal that can be discretized.

    You can use this class to create and store signals that will be sent to
    third-party devices. The signals created in this way are independent of
    Boulder Opal graphs and have a fixed time step between their segments.

    Parameters
    ----------
    function : Callable
        A function that returns the value of the signal at each instant of time.
        It must be capable of accepting a NumPy array of times as an input
        parameters, in which case it should return the values of the function
        for all the times passed.
    duration : float
        The duration of the signal.
    """

    def __init__(self, function: Callable, duration: float):
        check_argument(
            duration > 0, "The duration must be positive.", {"duration": duration}
        )

        self.duration = duration
        self._function = function

    def export_with_time_step(self, time_step: float) -> np.ndarray:
        """
        Return the values of the signal sampled at a constant rate given by the
        time step provided.

        Parameters
        ----------
        time_step : float
            The interval when the signal is to be sampled (that is, the duration
            of each segment of the discretized signal). It must be positive and
            shorter than the total duration of the signal.

        Returns
        -------
        np.ndarray
            An array with the values of the signal sampled at equal intervals.
            The value of the signal in each segment corresponds to the value of
            the function at the center of that segment.

        Warnings
        --------
        If the time step passed doesn't exactly divide the total duration of
        the signal, this function will round the number of segments of the
        discretized output to the nearest number that is an integer multiple of
        the time step.
        """
        check_argument(
            time_step > 0, "The time step must be positive.", {"time_step": time_step}
        )
        check_argument(
            self.duration >= time_step,
            "The time step must not be longer than the duration of the signal.",
            {"duration": self.duration, "time_step": time_step},
        )

        segment_count = int(np.round(self.duration / time_step))
        times = (np.arange(segment_count) + 0.5) * time_step

        return self._function(times)

    def export_with_sampling_rate(self, sampling_rate: float) -> np.ndarray:
        """
        Return the values of the signal sampled at a constant rate given by the
        sampling rate provided.

        Parameters
        ----------
        sampling_rate : float
            The rate at which the signal is sampled (that is, the inverse of the
            duration of each segment of the discretized signal). It must be
            positive and larger than the inverse of the duration.

        Returns
        -------
        np.ndarray
            An array with the values of the signal sampled at equal intervals.
            The value of the signal in each segment corresponds to the value of
            the function at the center of that segment.

        Warnings
        --------
        If the inverse of the sampling rate passed doesn't exactly divide the
        total duration of the signal, this function will round the number of
        segments of the discretized output to the nearest number that is an
        integer multiple of the inverse of the sampling rate.
        """
        check_argument(
            sampling_rate > 0,
            "The sampling rate must be positive.",
            {"sampling_rate": sampling_rate},
        )
        time_step = 1 / sampling_rate
        check_argument(
            self.duration >= time_step,
            "The inverse of the sampling rate must not be longer than the"
            " duration of the signal.",
            {"duration": self.duration, "sampling_rate": sampling_rate},
            extras={"1/sampling_rate": time_step},
        )

        segment_count = int(np.round(self.duration / time_step))
        times = (np.arange(segment_count) + 0.5) * time_step

        return self._function(times)


@expose(Namespace.SIGNALS)
class SegmentationType(Enum):
    """
    An enumeration of segmentation types for piecewise-constant signals.

    You can use this Enum to specify the segmentation type in the
    `graph nodes <https://docs.q-ctrl.com/boulder-opal/references/qctrl/Toolkits
    /signals.html#graph-nodes>`_ when defining a PWC signal containing a constant part.

    Attributes
    ----------
    UNIFORM
        A uniform segmentation distributes the piecewise-constant segments
        uniformly along the signal's duration.
        This segmentation type is more efficient when combining different signals,
        as their segmentations will match.

    MINIMAL
        A minimal segmentation represents constant parts of the piecewise-constant
        function with a minimal number of segments (typically one or two),
        reserving most of the segments for the non-constant parts.
        This segmentation type might be preferred when only a single signal is present
        in the system, as it leads to a more efficient sampling of the non-constant parts.
        However, combining signals with different segmentations can lead to uneven sampling
        and increased computation time due to the increase in number of segments.
    """

    UNIFORM = "UNIFORM"
    MINIMAL = "MINIMAL"


@expose(Namespace.SIGNALS)
@no_graph_parameters(alternatives="`graph.signals.square_pulse_pwc`")
def square_pulse(
    duration: float,
    amplitude: float | complex,
    start_time: float = 0.0,
    end_time: Optional[float] = None,
) -> Signal:
    r"""
    Create a `Signal` object representing a square pulse.

    Parameters
    ----------
    duration : float
         The duration of the pulse.
    amplitude : float or complex
        The amplitude of the square pulse, :math:`A`.
    start_time : float, optional
        The start time of the square pulse, :math:`t_0`.
        Defaults to 0.
    end_time : float, optional
        The end time of the square pulse, :math:`t_1`.
        Must be greater than the start time.
        Defaults to the `duration`.

    Returns
    -------
    Signal
        The square pulse.

    See Also
    --------
    :func:`.signals.cosine_pulse` : Create a `Signal` object representing a cosine pulse.
    :func:`.signals.gaussian_pulse` : Create a `Signal` object representing a Gaussian pulse.
    :func:`.signals.sech_pulse` : Create a `Signal` object representing a hyperbolic secant pulse.
    :func:`.signals.square_pulse_pwc` :
        Graph operation to create a `Pwc` representing a square pulse.

    Notes
    -----
    The square pulse is defined as

    .. math:: \mathop{\mathrm{Square}}(t) = A \theta(t-t_0) \theta(t_1-t) ,

    where :math:`\theta(t)` is the
    `Heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_.

    Examples
    --------
    Define a square pulse and discretize it.

    >>> pulse = qctrl.signals.square_pulse(
    ...     duration=4.0, amplitude=2.5, start_time=1.0, end_time=3.0
    ... )
    >>> pulse.export_with_time_step(time_step=1.0)
    array([0. , 2.5, 2.5, 0. ])
    """
    if end_time is None:
        end_time = duration

    check_argument(
        end_time > start_time,
        "The end time must be greater than the start time.",
        {"start_time": start_time, "end_time": end_time},
    )

    return Signal(
        function=lambda times: np.where(
            np.logical_and(times >= start_time, times <= end_time), amplitude, 0.0
        ),
        duration=duration,
    )


@expose(Namespace.SIGNALS)
@no_graph_parameters(alternatives="`graph.signals.cosine_pwc`")
def cosine_pulse(
    duration: float,
    amplitude: float | complex,
    drag: float = 0.0,
    start_time: float = 0.0,
    end_time: Optional[float] = None,
    flat_duration: float = 0.0,
) -> Signal:
    r"""
    Create a `Signal` object representing a cosine pulse.

    Parameters
    ----------
    duration : float
         The duration of the pulse.
    amplitude : float or complex
        The amplitude of the pulse, :math:`A`.
    drag : float, optional
        The DRAG parameter, :math:`\beta`.
        Defaults to 0, in which case there is no DRAG correction.
    start_time : float, optional
        The time at which the cosine pulse starts, :math:`t_\mathrm{start}`.
        Defaults to 0.
    end_time : float, optional
        The time at which the cosine pulse ends, :math:`t_\mathrm{end}`.
        Defaults to the `duration`.
    flat_duration : float, optional
        The amount of time that the pulse remains constant after the peak of
        the cosine, :math:`t_\mathrm{flat}`.
        If passed, it must be nonnegative and less than the difference between
        `end_time` and `start_time`.
        Defaults to 0, in which case no constant part is added to the cosine pulse.

    Returns
    -------
    Signal
        The cosine pulse.

    See Also
    --------
    :func:`.signals.cosine_pulse_pwc` :
        Graph operation to create a `Pwc` representing a cosine pulse.
    :func:`.signals.gaussian_pulse` : Create a `Signal` object representing a Gaussian pulse.
    :func:`.signals.hann_series` :
        Create a `Signal` object representing a sum of Hann window functions.
    :func:`.signals.sech_pulse` : Create a `Signal` object representing a hyperbolic secant pulse.
    :func:`.signals.sinusoid` : Create a `Signal` object representing a sinusoidal oscillation.
    :func:`.signals.square_pulse` : Create a `Signal` object representing a square pulse.

    Notes
    -----
    The cosine pulse is defined as

    .. math:: \mathop{\mathrm{Cos}}(t) =
        \begin{cases}
        0
        &\mathrm{if} \quad t < t_\mathrm{start} \\
        \frac{A}{2} \left[1+\cos \left(\omega \{t-\tau_-\} \right)
        + i\omega\beta \sin \left(\omega \{t-\tau_-\}\right)\right]
        &\mathrm{if} \quad t_\mathrm{start} \le t < \tau_- \\
        A
        &\mathrm{if} \quad \tau_- \le t \le \tau_+ \\
        \frac{A}{2} \left[1+\cos \left(\omega\{t-\tau_+\}\right)
        + i\omega \beta\sin \left(\omega \{t-\tau_+\}\right)\right]
        &\mathrm{if} \quad \tau_+ < t \le t_\mathrm{end} \\
        0
        &\mathrm{if} \quad t > t_\mathrm{end} \\
        \end{cases},

    where :math:`\omega=2\pi /(t_\mathrm{end}-t_\mathrm{start} - t_\mathrm{flat})`,
    :math:`\tau_\mp` are the start/end times of the flat segment,
    with :math:`\tau_\mp=(t_\mathrm{start}+t_\mathrm{end} \mp t_\mathrm{flat})/2`.

    If the flat duration is zero (the default setting), this reduces to

    .. math:: \mathop{\mathrm{Cos}}(t) =
        \frac{A}{2} \left[1+\cos \left(\omega \{t-\tau\} \right)
        + i\omega\beta \sin \left(\omega \{t-\tau\}\right)\right]
        \theta(t-t_\mathrm{start}) \theta(t_\mathrm{end}-t),

    where now :math:`\omega=2\pi /(t_\mathrm{end}-t_\mathrm{start})`,
    :math:`\tau=(t_\mathrm{start}+t_\mathrm{end})/2`
    and :math:`\theta(t)` is the
    `Heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_.

    Examples
    --------
    Define a cosine pulse.

    >>> pulse = qctrl.signals.cosine_pulse(duration=3.0, amplitude=1.0)
    >>> pulse.export_with_time_step(time_step=0.5)
    array([0.0669873+0.j, 0.5      +0.j, 0.9330127+0.j, 0.9330127+0.j,
           0.5      +0.j, 0.0669873+0.j])

    Define a flat-top cosine pulse with a DRAG correction.

    >>> pulse = qctrl.signals.cosine_pulse(
    ...     duration=3.0, amplitude=1.0, drag=0.1, flat_duration=0.6
    ... )
    >>> pulse.export_with_sampling_rate(sampling_rate=2.0)
    array([0.10332333-0.07968668j, 0.69134172-0.12093555j,
           1.        +0.j        , 1.        +0.j        ,
           0.69134172+0.12093555j, 0.10332333+0.07968668j])
    """
    if end_time is None:
        end_time = duration

    check_argument(
        end_time > start_time,
        "The end time must be greater than the start time.",
        {"start_time": start_time, "end_time": end_time},
    )
    check_argument(
        0 <= flat_duration <= (end_time - start_time),
        "The duration of the flat part of the pulse has to be nonnegative and"
        "smaller than or equal to the total duration of the pulse.",
        {
            "flat_duration": flat_duration,
            "start_time": start_time,
            "end_time": end_time,
        },
    )

    pulse_period = end_time - start_time - flat_duration

    flat_segment_start = start_time + 0.5 * pulse_period
    flat_segment_end = end_time - 0.5 * pulse_period

    angular_frequency = 2.0 * np.pi / pulse_period

    def _cosine_pulse(times: np.ndarray) -> np.ndarray:
        shifted_times = np.where(
            times < flat_segment_start,
            times - flat_segment_start,
            times - flat_segment_end,
        )
        values = (0.5 * amplitude) * (
            1
            + np.cos(angular_frequency * shifted_times)
            + (angular_frequency * drag * 1j)
            * np.sin(angular_frequency * shifted_times)
        )

        # Make pulse flat for the duration of the "flat segment".
        flat_values = np.where(
            np.logical_and(times > flat_segment_start, times < flat_segment_end),
            amplitude,
            values,
        )

        # Make the pulse zero before its start and after its end.
        limited_values = np.where(
            np.logical_and(times > start_time, times < end_time), flat_values, 0
        )
        return limited_values

    return Signal(function=_cosine_pulse, duration=duration)


@expose(Namespace.SIGNALS)
@no_graph_parameters(
    alternatives="`graph.signals.sinusoid_pwc` or `graph.signals.sinusoid_stf`"
)
def sinusoid(
    duration: float,
    amplitude: float | complex,
    angular_frequency: float,
    phase: float = 0.0,
) -> Signal:
    r"""
    Create a `Signal` object representing a sinusoidal oscillation.

    Parameters
    ----------
    duration : float
        The duration of the oscillation.
    amplitude : float or complex
        The amplitude of the oscillation, :math:`A`.
    angular_frequency : float
        The angular frequency of the oscillation, :math:`\omega`.
    phase : float, optional
        The phase of the oscillation, :math:`\phi`.
        Defaults to 0.

    Returns
    -------
    Signal
        The sinusoidal oscillation.

    See Also
    --------
    :func:`.signals.cosine_pulse` : Create a `Signal` object representing a cosine pulse.
    :func:`.signals.hann_series` :
        Create a `Signal` object representing a sum of Hann window functions.
    :func:`.signals.sinusoid_pwc` :
        Graph operation to create a `Pwc` representing a sinusoidal oscillation.
    :func:`.signals.sinusoid_stf` :
        Graph operation to create a `Stf` representing a sinusoidal oscillation.

    Notes
    -----
    The sinusoid is defined as

    .. math:: \mathop{\mathrm{Sinusoid}}(t) = A \sin \left( \omega t + \phi \right) .

    Examples
    --------
    Define a sinusoidal oscillation.

    >>> signal = qctrl.signals.sinusoid(
    ...     duration=2.0,
    ...     amplitude=1.0,
    ...     angular_frequency=np.pi,
    ...     phase=np.pi/2.0,
    ... )
    >>> signal.export_with_sampling_rate(sampling_rate=0.25)
    array([ 0.92387953,  0.38268343, -0.38268343, -0.92387953, -0.92387953,
       -0.38268343,  0.38268343,  0.92387953])
    """

    return Signal(
        function=lambda times: amplitude * np.sin(angular_frequency * times + phase),
        duration=duration,
    )


@expose(Namespace.SIGNALS)
@no_graph_parameters(
    alternatives="`graph.signals.hann_series_pwc` or `graph.signals.hann_series_stf`"
)
def hann_series(duration: float, coefficients: np.ndarray) -> Signal:
    r"""
    Create a `Signal` object representing a sum of Hann window functions.

    Parameters
    ----------
    duration : float
        The duration of the signal, :math:`T`.
    coefficients : np.ndarray
        The coefficients for the different Hann window functions, :math:`c_n`.
        It must be a 1D array.

    Returns
    -------
    Signal
        The Hann window functions series.

    See Also
    --------
    :func:`.signals.cosine_pulse` : Create a `Signal` object representing a cosine pulse.
    :func:`.signals.sinusoid` : Create a `Signal` object representing a sinusoidal oscillation.
    :func:`.signals.hann_series_pwc` :
        Graph operation to create a `Pwc` representing a sum of Hann window functions.
    :func:`.signals.hann_series_stf` :
        Graph operation to create an `Stf` representing a sum of Hann window functions.

    Notes
    -----
    The series is defined as

    .. math:: \mathop{\mathrm{Hann}}(t)
        = \sum_{n=1}^N c_n \sin^2 \left( \frac{\pi n t}{T} \right) ,

    where :math:`N` is the number of coefficients.

    Examples
    --------
    Define a simple Hann series.

    >>> signal = graph.signals.hann_series(
    ...     duration=5.0,
    ...     coefficients=np.array([0.5, 1, 0.25]),
    ... )
    >>> signal.export_with_time_step(time_step=0.5)
    array([0.15925422, 1.00144425, 1.375     , 1.05757275, 0.78172879,
       0.78172879, 1.05757275, 1.375     , 1.00144425, 0.15925422])
    """

    check_argument(
        len(coefficients.shape) == 1,
        "The coefficients must be in a 1D array.",
        {"coefficients": coefficients},
        extras={"coefficients.shape": coefficients.shape},
    )

    nss = np.arange(1, coefficients.shape[0] + 1)

    return Signal(
        function=lambda times: np.sum(
            coefficients * np.sin(np.pi * nss * times[:, None] / duration) ** 2, axis=1
        ),
        duration=duration,
    )


@expose(Namespace.SIGNALS)
@no_graph_parameters(
    alternatives="`graph.signals.linear_ramp_pwc` or `graph.signals.linear_ramp_stf`"
)
def linear_ramp(
    duration: float,
    end_value: float | complex,
    start_value: Optional[float | complex] = None,
    start_time: float = 0.0,
    end_time: Optional[float] = None,
) -> Signal:
    r"""
    Create a `Signal` object representing a linear ramp.

    Parameters
    ----------
    duration : float
        The duration of the signal, :math:`T`.
    end_value : float or complex
        The value of the ramp at :math:`t = t_\mathrm{end}`, :math:`a_\mathrm{end}`.
    start_value : float or complex, optional
        The value of the ramp at :math:`t = t_\mathrm{start}`, :math:`a_\mathrm{start}`.
        Defaults to :math:`-a_\mathrm{end}`.
    start_time : float, optional
        The time at which the linear ramp starts, :math:`t_\mathrm{start}`.
        Defaults to 0.
    end_time : float, optional
        The time at which the linear ramp ends, :math:`t_\mathrm{end}`.
        Defaults to the given duration :math:`T`.

    Returns
    -------
    Signal
        The linear ramp.

    See Also
    --------
    :func:`.signals.linear_ramp_pwc` : Graph operation to create a `Pwc` representing a linear ramp.
    :func:`.signals.linear_ramp_stf` : Graph operation to create a `Stf` representing a linear ramp.
    :func:`.signals.tanh_ramp` : Create a `Signal` object representing a hyperbolic tangent ramp.

    Notes
    -----
    The linear ramp is defined as

    .. math:: \mathop{\mathrm{Linear}}(t) =
        \begin{cases}
            a_\mathrm{start} &\mathrm{if} \quad t < t_\mathrm{start}\\
            a_\mathrm{start} + (a_\mathrm{end} - a_\mathrm{start})
                \frac{t - t_\mathrm{start}}{t_\mathrm{end} - t_\mathrm{start}}
                &\mathrm{if} \quad t_\mathrm{start} \le t \le t_\mathrm{end} \\
            a_\mathrm{end} &\mathrm{if} \quad t > t_\mathrm{end}
        \end{cases} .

    Examples
    --------
    Define a linear ramp with start and end times.

    >>> signal = qctrl.signals.linear_ramp(
    ...     duration=4, end_value=2, start_time=1, end_time=3
    ... )
    >>> signal.export_with_time_step(time_step=0.25)
    array([-2.  , -2.  , -2.  , -2.  , -1.75, -1.25, -0.75, -0.25,  0.25,
        0.75,  1.25,  1.75,  2.  ,  2.  ,  2.  ,  2.  ])
    """
    if start_value is None:
        start_value = -end_value

    if end_time is None:
        end_time = duration

    check_argument(
        start_time < end_time,
        "The end time of the pulse must be greater than the start time.",
        {"end_time": end_time, "start_time": start_time},
    )

    slope = (end_value - start_value) / (end_time - start_time)

    def _linear_ramp(times: np.ndarray) -> np.ndarray:
        assert start_value is not None  # make mypy happy
        result = np.where(times > end_time, end_value - start_value, 0.0)
        result += start_value + np.where(
            np.logical_and(times >= start_time, times <= end_time),
            slope * (times - start_time),
            0,
        )
        return result

    return Signal(function=_linear_ramp, duration=duration)


@expose(Namespace.SIGNALS)
@no_graph_parameters(
    alternatives="`graph.signals.tanh_ramp_pwc` or `graph.signals.tanh_ramp_stf`"
)
def tanh_ramp(
    duration: float,
    end_value: float | complex,
    start_value: Optional[float | complex] = None,
    ramp_duration: Optional[float] = None,
    center_time: Optional[float] = None,
) -> Signal:
    r"""
    Create a `Signal` object representing a hyperbolic tangent ramp.

    Parameters
    ----------
    duration : float
        The duration of the signal, :math:`T`.
    end_value : float or complex
        The asymptotic value of the ramp towards :math:`t \to +\infty`, :math:`a_+`.
    start_value : float or complex, optional
        The asymptotic value of the ramp towards :math:`t \to -\infty`, :math:`a_-`.
        Defaults to minus `end_value`.
    ramp_duration : float, optional
        The characteristic time for the hyperbolic tangent ramp, :math:`t_\mathrm{ramp}`.
        Defaults to :math:`T/6`.
    center_time : float, optional
        The time at which the ramp has its greatest slope, :math:`t_0`.
        Defaults to :math:`T/2`.

    Returns
    -------
    Signal
        The hyperbolic tangent ramp.

    See Also
    --------
    :func:`.signals.linear_ramp` : Create a `Signal` object representing a linear ramp.
    :func:`.signals.tanh_ramp_pwc` :
        Graph operation to create a `Pwc` representing a hyperbolic tangent ramp.
    :func:`.signals.tanh_ramp_stf` :
        Graph operation to create a `Stf` representing a hyperbolic tangent ramp.

    Notes
    -----
    The hyperbolic tangent ramp is defined as

    .. math:: \mathop{\mathrm{Tanh}}(t)
        = \frac{a_+ + a_-}{2}
            + \frac{a_+ - a_-}{2} \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) ,

    where the function's asymptotic values :math:`a_\pm` are defined by:

    .. math::  a_\pm := \lim_{t\to\pm\infty} \mathop{\mathrm{Tanh}}(t) ,

    and :math:`t_0` is related to :math:`t_\mathrm{ramp}` by:

    .. math::
        \left.\frac{{\rm d}\mathop{\mathrm{Tanh}}(t)}{{\rm d}t}\right|_{t=t_0}
            = \frac{ (a_+ - a_-)}{2 t_\mathrm{ramp}} .

    Note that if :math:`t_0` is close to the edges of the ramp,
    for example :math:`t_0 \lesssim 2 t_\mathrm{ramp}`,
    then the first and last values of the outputted array will differ from the
    expected asymptotic values.

    With the default values of `start_value` (:math:`a_-`),
    `ramp_duration` (:math:`t_\mathrm{ramp}`), and `center_time` (:math:`t_0`),
    the ramp expression simplifies to

    .. math:: \mathop{\mathrm{Tanh}}(t) = A \tanh\left( \frac{t - T/2}{T/6} \right),

    where :math:`A = a_+` is the end value (the start value is then :math:`-A`).
    This defines a symmetric ramp (around :math:`(T/2, 0)`)
    between :math:`-0.995 A` (at :math:`t=0`) and :math:`0.995 A` (at :math:`t=T`).

    Examples
    --------
    Define a tanh ramp.

    >>> signal = qctrl.signals.tanh_ramp(
    ...     duration=4, end_value=2, start_value=1, ramp_duration=0.4, center_time=2.
    ... )
    >>> signal.export_with_time_step(time_step=0.4)
    array([1.00012339, 1.00091105, 1.00669285, 1.04742587, 1.26894142,
       1.73105858, 1.95257413, 1.99330715, 1.99908895, 1.99987661])
    """
    if start_value is None:
        start_value = -end_value

    if ramp_duration is None:
        ramp_duration = duration / 6

    if center_time is None:
        center_time = duration / 2

    def _tanh_ramp(times):
        assert start_value is not None  # make mypy happy
        return start_value + 0.5 * (end_value - start_value) * (
            1 + np.tanh((times - center_time) / ramp_duration)
        )

    return Signal(function=_tanh_ramp, duration=duration)


@expose(Namespace.SIGNALS)
@no_graph_parameters(
    alternatives="`graph.signals.sech_pulse_pwc` or `graph.signals.sech_pulse_stf`"
)
def sech_pulse(
    duration: float,
    amplitude: float | complex,
    width: Optional[float] = None,
    center_time: Optional[float] = None,
) -> Signal:
    r"""
    Create a `Signal` object representing a hyperbolic secant pulse.

    Parameters
    ----------
    duration : float
        The duration of the signal, :math:`T`.
    amplitude : float or complex
        The amplitude of the pulse, :math:`A`.
    width : float, optional
        The characteristic time for the hyperbolic secant pulse, :math:`t_\mathrm{pulse}`.
        Defaults to :math:`T/12`,
        giving the pulse a full width at half maximum (FWHM) of :math:`0.22 T`.
    center_time : float, optional
        The time at which the pulse peaks, :math:`t_\mathrm{peak}`.
        Defaults to :math:`T/2`.

    Returns
    -------
    Signal
        The hyperbolic secant pulse.

    See Also
    --------
    :func:`.signals.cosine_pulse` : Create a `Signal` object representing a cosine pulse.
    :func:`.signals.gaussian_pulse` : Create a `Signal` object representing a Gaussian pulse.
    :func:`.signals.sech_pulse_pwc` :
        Graph operation to create a `Pwc` representing a hyperbolic secant pulse.
    :func:`.signals.sech_pulse_stf` :
        Graph operation to create a `Stf` representing a hyperbolic secant pulse.
    :func:`.signals.square_pulse` : Create a `Signal` object representing a square pulse.

    Notes
    -----
    The hyperbolic secant pulse is defined as

    .. math:: \mathop{\mathrm{Sech}}(t)
        = \frac{A}{\cosh\left((t - t_\mathrm{peak}) / t_\mathrm{pulse} \right)} .

    The FWHM of the pulse is about :math:`2.634 t_\mathrm{pulse}`.

    Examples
    --------
    Define a displaced sech pulse.

    >>> pulse = qctrl.signals.sech_pulse(duration=1, amplitude=1, center_time=0.4)
    >>> pulse.export_with_time_step(time_step=0.1)
    array([0.02998441, 0.09932793, 0.32180487, 0.84355069, 0.84355069,
       0.32180487, 0.09932793, 0.02998441, 0.00903298, 0.00272073])
    """
    if width is None:
        width = duration / 12

    if center_time is None:
        center_time = duration / 2

    return Signal(
        function=lambda times: amplitude / np.cosh((times - center_time) / width),
        duration=duration,
    )


@expose(Namespace.SIGNALS)
@no_graph_parameters(
    alternatives="`graph.signals.gaussian_pulse_pwc` or `graph.signals.gaussian_pulse_stf`"
)
def gaussian_pulse(
    duration: float,
    amplitude: float | complex,
    width: Optional[float] = None,
    center_time: Optional[float] = None,
    drag: float = 0.0,
    flat_duration: float = 0.0,
) -> Signal:
    r"""
    Create a `Signal` object representing a Gaussian pulse.

    Parameters
    ----------
    duration : float
        The duration of the signal, :math:`T`.
    amplitude : float or complex
        The amplitude of the Gaussian pulse, :math:`A`.
    width : float, optional
        The standard deviation of the Gaussian pulse, :math:`\sigma`.
        Defaults to :math:`T/10` or :math:`(T-t_\mathrm{flat})/10` if `flat_duration` is passed.
    center_time : float, optional
        The center of the Gaussian pulse, :math:`t_0`.
        Defaults to half of the given value of the duration, :math:`T/2`.
    drag : float, optional
        The DRAG parameter, :math:`\beta`.
        Defaults to 0, in which case there is no DRAG correction.
    flat_duration : float, optional
        The amount of time to remain constant after the peak of the Gaussian,
        :math:`t_\mathrm{flat}`.
        If passed, it must be nonnegative and less than the duration.
        Defaults to 0, in which case no constant part is added to the Gaussian pulse.

    Returns
    -------
    Signal
        The Gaussian pulse.

    See Also
    --------
    :func:`.signals.cosine_pulse` : Create a `Signal` object representing a cosine pulse.
    :func:`.signals.gaussian_pulse_pwc` :
        Graph operation to create a `Pwc` representing a Gaussian pulse.
    :func:`.signals.gaussian_pulse_stf` :
        Graph operation to create a `Stf` representing a Gaussian pulse.
    :func:`.signals.sech_pulse` : Create a `Signal` object representing a hyperbolic secant pulse.
    :func:`.signals.square_pulse` : Create a `Signal` object representing a square pulse.

    Notes
    -----
    The Gaussian pulse is defined as

    .. math:: \mathop{\mathrm{Gaussian}}(t) =
        \begin{cases}
            A \left(1-\frac{i\beta (t-t_1)}{\sigma^2}\right)
            \exp \left(- \frac{(t-t_1)^2}{2\sigma^2} \right)
                &\mathrm{if} \quad t < t_1=t_0- t_\mathrm{flat}/2\\
            A
                &\mathrm{if} \quad t_0-t_\mathrm{flat}/2 \le t < t_0+t_\mathrm{flat}/2 \\
            A \left(1-\frac{i\beta (t-t_2)}{\sigma^2}\right)
            \exp \left(- \frac{(t-t_2)^2}{2\sigma^2} \right)
                &\mathrm{if} \quad t > t_2=t_0+t_\mathrm{flat}/2
        \end{cases} .

    If the flat duration is zero (the default setting), this reduces to

    .. math:: \mathop{\mathrm{Gaussian}}(t) =
        A \left(1-\frac{i\beta (t-t_0)}{\sigma^2}\right)
        \exp \left(- \frac{(t-t_0)^2}{2\sigma^2} \right) .

    Examples
    --------
    Define a Gaussian pulse.

    >>> pulse = qctrl.signals.gaussian_pulse(duration=2.0, amplitude=1.0)
    >>> pulse.export_with_time_step(time_step=0.2)
    array([4.00652974e-05+0.j, 2.18749112e-03+0.j, 4.39369336e-02+0.j,
       3.24652467e-01+0.j, 8.82496903e-01+0.j, 8.82496903e-01+0.j,
       3.24652467e-01+0.j, 4.39369336e-02+0.j, 2.18749112e-03+0.j,
       4.00652974e-05+0.j])

    Define a flat-top Gaussian pulse with a DRAG correction.

    >>> pulse = qctrl.signals.gaussian_pulse(
    ...     duration=1.0,
    ...     amplitude=1.0,
    ...     width=0.2,
    ...     center_time=0.6,
    ...     drag=0.1,
    ...     flat_duration=0.2,
    ... )
    >>> pulse.export_with_sampling_rate(sampling_rate=10.)
    array([0.07955951+0.08950445j, 0.21626517+0.18923202j,
       0.45783336+0.28614585j, 0.7548396 +0.28306485j,
       0.96923323+0.12115415j, 1.        +0.j        ,
       1.        +0.j        , 0.96923323-0.12115415j,
       0.7548396 -0.28306485j, 0.45783336-0.28614585j])
    """

    check_argument(
        0.0 <= flat_duration < duration,
        "The flat duration must be nonnegative and less than the duration.",
        {"flat_duration": flat_duration},
        extras={"duration": duration},
    )

    if center_time is None:
        center_time = 0.5 * duration

    if width is None:
        if flat_duration is None:
            width = duration / 10
        else:
            width = (duration - flat_duration) / 10

    # Time at which first Gaussian ends.
    flat_segment_start = center_time - 0.5 * flat_duration
    # Time at which second Gaussian starts.
    flat_segment_end = center_time + 0.5 * flat_duration

    def _gaussian_pulse(times: np.ndarray) -> np.ndarray:
        assert width is not None  # make mypy happy

        shifted_times = np.where(
            times < flat_segment_start,
            times - flat_segment_start,
            times - flat_segment_end,
        )

        values = amplitude * (1.0 - 1j * drag * shifted_times / (width**2))
        values *= np.exp(-(shifted_times**2) / (2 * width**2))

        # Make pulse flat for the duration of the "flat segment".
        flat_values = np.where(
            np.logical_and(times > flat_segment_start, times < flat_segment_end),
            amplitude,
            values,
        )

        return flat_values

    return Signal(function=_gaussian_pulse, duration=duration)
